#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.stopwatch import StopWatch
from datetime import timedelta
import functools
from multiprocessing import cpu_count, Pool
import itertools
from covsirphy.util.error import UnExecutedError, deprecate
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.ode_solver_multi import _MultiPhaseODESolver
from covsirphy._deprecated.param_estimator import _ParamEstimator


class ODEHandler(Term):
    """
    Perform simulation and parameter estimation with a multi-phased ODE model.

    Args:
        model (covsirphy.ModelBase): ODE model
        first_date (str or pandas.Timestamp): the first date of simulation, like 14Apr2021
        tau (int or None): tau value [min] or None (to be estimated)
        metric (str): metric name for estimation
        n_jobs (int): the number of parallel jobs or -1 (CPU count)
    """

    @deprecate(old="ODEHandler", new="Dynamics", version="2.24.0-xi")
    def __init__(self, model, first_date, tau=None, metric="RMSLE", n_jobs=-1):
        self._model = Validator(model, "model").subclass(ModelBase)
        self._first = Validator(first_date, "first_date").date()
        self._metric = Validator([metric], "metric").sequence(candidates=Evaluator.metrics())[0]
        self._n_jobs = cpu_count() if n_jobs == -1 else Validator(n_jobs, "n_jobs").int()
        # Tau value [min] or None
        self._tau = Validator(tau, "tau").tau(default=None)
        # {"0th": output of self.add()}
        self._info_dict = {}

    def add(self, end_date, param_dict=None, y0_dict=None):
        """
        Add a new phase.

        Args:
            end_date (str or pandas.Timestamp): end date of the phase
            param_dict (dict[str, float] or None): parameter values or None (not set)
            y0_dict (dict[str, int] or None): initial values or None (not set)

        Returns:
            dict(str, object): setting of the phase (key: phase name)
                - Start (pandas.Timestamp): start date
                - End (pandas.Timestamp): end date
                - y0 (dict[str, int]): initial values of model-specialized variables or empty dict
                - param (dict[str, float]): parameter values or empty dict
        """
        if not self._info_dict and y0_dict is None:
            raise ValueError("@y0_dict must be specified for the 0th phase, but None was applied.")
        phase = self.num2str(len(self._info_dict))
        if self._info_dict:
            start = list(self._info_dict.values())[-1][self.END] + timedelta(days=1)
        else:
            start = self._first
        end = Validator(end_date, "end_date").date()
        self._info_dict[phase] = {
            self.START: start, self.END: end, "y0": y0_dict or {}, "param": param_dict or {}}
        return self._info_dict[phase]

    def simulate(self):
        """
        Perform simulation with the multi-phased ODE model.

        Raises:
            covsirphy.UnExecutedError: either tau value or phase information was not set

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        if self._tau is None:
            raise UnExecutedError(
                "ODEHandler.estimate_tau()", details="Or specify tau when creating an instance of ODEHandler")
        if not self._info_dict:
            raise UnExecutedError("ODEHandler.add()")
        combs = itertools.product(self._model.PARAMETERS, self._info_dict.items())
        for (param, (phase, phase_dict)) in combs:
            if param not in phase_dict["param"]:
                raise ValueError(f"{param.capitalize()} is not registered for the {phase} phase.")
        solver = _MultiPhaseODESolver(self._model, self._first, self._tau)
        return solver.simulate(*self._info_dict.values())

    def _score_tau(self, tau, data, quantile):
        """
        Calculate score for the tau value.

        Args:
            tau (int): tau value [min]
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            quantile (float): quantile to guess ODE parameter values for the candidates of tau
        """
        info_dict = self._info_dict.copy()
        for (phase, phase_dict) in info_dict.items():
            start, end = phase_dict[self.START], phase_dict[self.END]
            df = data.loc[(start <= data[self.DATE]) & (data[self.DATE] <= end)]
            info_dict[phase]["param"] = self._model.guess(df, tau, q=quantile)
        solver = _MultiPhaseODESolver(self._model, self._first, tau)
        sim_df = solver.simulate(*info_dict.values())
        evaluator = Evaluator(data.set_index(self.DATE), sim_df.set_index(self.DATE))
        return evaluator.score(metric=self._metric)

    def estimate_tau(self, data, guess_quantile=0.5):
        """
        Select tau value [min] which minimize the score of the metric.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            guess_quantile (float): quantile to guess ODE parameter values for the candidates of tau

        Returns:
            int: estimated tau value [min]

        Raises:
            covsirphy.UnExecutedError: phase information was not set

        Note:
            ODE parameter for each tau value will be guessed by .guess() classmethod of the model.
            Tau value will be selected from the divisors of 1440 [min] and set to self.
        """
        Validator(data, "data").dataframe(columns=self.DSIFR_COLUMNS)
        df = data.loc[:, self.DSIFR_COLUMNS]
        if not self._info_dict:
            raise UnExecutedError("ODEHandler.add()")
        # Calculate scores of tau candidates
        Validator(guess_quantile, "quantile").float(value_range=(0, 1))
        calc_f = functools.partial(self._score_tau, data=df, quantile=guess_quantile)
        divisors = [i for i in range(1, 1441) if 1440 % i == 0]
        if self._n_jobs == 1:
            scores = [calc_f(candidate) for candidate in divisors]
        else:
            with Pool(self._n_jobs) as p:
                scores = p.map(calc_f, divisors)
        score_dict = dict(zip(divisors, scores))
        # Return the best tau value
        comp_f = {True: min, False: max}[Evaluator.smaller_is_better(metric=self._metric)]
        self._tau = comp_f(score_dict.items(), key=lambda x: x[1])[0]
        return self._tau

    def _estimate_params(self, phase, data, quantiles, check_dict, study_dict, show_phase):
        """
        Perform parameter estimation for one phase.

        Args:
            phase (str): phase name
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            quantiles (tuple(int, int)): quantiles to cut parameter range, like confidence interval
            check_dict (dict[str, object]): setting of validation
            study_dict (dict[str, object]): setting of optimization study
            show_phase (bool): whether show phase name or not (stdout)

        Returns:
            dict(str, object):
                - Rt (float): phase-dependent reproduction number
                - (str, float): estimated parameter values, including rho
                - (int or float): day parameters, including 1/beta [days]
                - {metric}: score with the estimated parameter values
                - Trials (int): the number of trials
                - Runtime (str): runtime of optimization
        """
        phase_dict = self._info_dict[phase].copy()
        start, end = phase_dict[self.START], phase_dict[self.END]
        df = data.loc[(start <= data[self.DATE]) & (data[self.DATE] <= end)]
        estimator = _ParamEstimator(self._model, df, self._tau, self._metric, quantiles)
        est_dict = estimator.run(check_dict, study_dict)
        n_trials, runtime = est_dict[self.TRIALS], est_dict[self.RUNTIME]
        start_date = start.strftime(self.DATE_FORMAT)
        end_date = end.strftime(self.DATE_FORMAT)
        if show_phase:
            ph_statement = f"{phase:>4} phase ({start_date} - {end_date})"
        else:
            ph_statement = f"{start_date} - {end_date}"
        print(f"\t{ph_statement}: finished {n_trials:>4} trials in {runtime}")
        return est_dict

    def estimate_params(self, data, quantiles=(0.1, 0.9), check_dict=None, study_dict=None, **kwargs):
        """
        Estimate ODE parameter values of the all phases to minimize the score of the metric.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            quantiles (tuple(int, int)): quantiles to cut parameter range, like confidence interval
            check_dict (dict[str, object] or None): setting of validation
                - None means {"timeout": 180, "timeout_interaction": 1, "tail_n": 4, "allowance": (0.99, 1.01)}
                - timeout (int): timeout of optimization
                - timeout_iteration (int): timeout of one iteration
                - tail_n (int): the number of iterations to decide whether score did not change for the last iterations
                - allowance (tuple(float, float)): the allowance of the max predicted values
            study_dict (dict[str, object] or None): setting of optimization study
                - None means {"pruner": "threshold", "upper": 0.5, "percentile": 50, "seed": 0, "constant_liar": False}
                - pruner (str): kind of pruner (hyperband, median, threshold or percentile)
                - upper (float): works for "threshold" pruner, intermediate score is larger than this value, it prunes
                - percentile (float): works for "Percentile" pruner, the best intermediate value is in the bottom percentile among trials, it prunes
                - constant_liar (bool): whether use constant liar to reduce search effort or not
            kwargs: we can set arguments directly. E.g. timeout=180 for check_dict={"timeout": 180,...}

        Raises:
            covsirphy.UnExecutedError: either tau value or phase information was not set

        Returns:
            dict(str, object): setting of the phase (key: phase name)
                - Start (pandas.Timestamp): start date
                - End (pandas.Timestamp): end date
                - Rt (float): phase-dependent reproduction number
                - (str, float): estimated parameter values, including rho
                - (int or float): day parameters, including 1/beta [days]
                - {metric}: score with the estimated parameter values
                - Trials (int): the number of trials
                - Runtime (str): runtime of optimization
        """
        print(f"\n<{self._model.NAME} model: parameter estimation>")
        print(f"Running optimization with {self._n_jobs} CPUs...")
        stopwatch = StopWatch()
        # Arguments
        Validator(data, "data").dataframe(columns=self.DSIFR_COLUMNS)
        df = data.loc[:, self.DSIFR_COLUMNS]
        if not self._info_dict:
            raise UnExecutedError("ODEHandler.add()")
        if self._tau is None:
            raise UnExecutedError(
                "ODEHandler.estimate_tau()",
                details="Or specify tau when creating an instance of ODEHandler")
        # Arguments of _ParamEstimator
        check_kwargs = {"timeout": 180, "timeout_iteration": 1, "tail_n": 4, "allowance": (0.99, 1.01)}
        check_kwargs.update(check_dict or {})
        check_kwargs.update(kwargs)
        study_kwargs = {
            "pruner": "threshold", "upper": 0.5, "percentile": 50, "seed": 0, "constant_liar": False}
        study_kwargs.update(study_dict or {})
        study_kwargs.update(kwargs)
        # ODE parameter estimation
        phases = list(self._info_dict.keys())
        est_f = functools.partial(
            self._estimate_params, data=df, quantiles=quantiles,
            check_dict=check_kwargs, study_dict=study_kwargs, show_phase=(len(phases) > 1))
        if self._n_jobs == 1:
            est_dict_list = [est_f(ph) for ph in phases]
        else:
            with Pool(self._n_jobs) as p:
                est_dict_list = p.map(est_f, phases)
        for (phase, est_dict) in zip(phases, est_dict_list):
            self._info_dict[phase]["param"] = {
                param: est_dict[param] for param in self._model.PARAMETERS}
        print(f"Completed optimization. Total: {stopwatch.stop_show()}")
        return {
            k: {self.START: self._info_dict[k][self.START], self.END: self._info_dict[k][self.END], **v}
            for (k, v) in zip(phases, est_dict_list)}

    def estimate(self, data, **kwargs):
        """
        Estimate tau value [min] and ODE parameter values.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            kwargs: keyword arguments of  ODEHandler.estimate_tau() and ODEHandler.estimate_param()

        Raises:
            covsirphy.UnExecutedError: phase information was not set

        Returns:
            tuple(int, dict(str, dict[str, object]))
                - int: tau value [min]
                - dict(str, object): setting of the phase (key: phase name)
                    - Start (pandas.Timestamp): start date
                    - End (pandas.Timestamp): end date
                    - Rt (float): phase-dependent reproduction number
                    - (str, float): estimated parameter values, including rho
                    - (int or float): day parameters, including 1/beta [days]
                    - {metric}: score with the estimated parameter values
                    - Trials (int): the number of trials
                    - Runtime (str): runtime of optimization
        """
        tau = self.estimate_tau(data, **Validator(kwargs, "keyword arguments").kwargs(functions=self.estimate_tau))
        info_dict = self.estimate_params(data, **kwargs)
        return (tau, info_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from covsirphy.util.filer import Filer
from covsirphy.util.term import Term
from covsirphy.downloading.provider import _DataProvider


class _DataBase(Term):
    """Basic class for databases.

    Args:
        directory (str or pathlib.Path): directory to save downloaded datasets
        update_interval (int): update interval of downloading dataset
        verbose (int): level of verbosity when downloading

    Note:
        If @verbose is 0, no descriptions will be shown.
        If @verbose is 1 or larger, URL and database name will be shown.
    """
    # File title without extensions and suffix
    TITLE = ""
    # Dictionary of column names
    COL_DICT = {}
    # Stdout when downloading (shown at most one time)
    STDOUT = None
    # Citation
    CITATION = ""

    def __init__(self, directory, update_interval, verbose):
        self._filer = Filer(directory=directory)
        self._update_interval = update_interval
        self._stdout = self.STDOUT if verbose > 0 else None
        self._provider = _DataProvider(update_interval=self._update_interval, stdout=self._stdout)

    def layer(self, country, province):
        """Return the data at the selected layer.

        Args:
            country (str or None): country name or None
            province (str or None): province/state/prefecture name or None

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - the other available columns

        Note:
            When @country is None, country-level data will be returned.

        Note:
            When @country is a string and @province is None, province-level data in the country will be returned.

        Note:
            When @country and @province are strings, city-level data in the province will be returned.
        """
        if country is None:
            return self._country()
        if province is None:
            return self._province(country=country)
        return self._city(country=country, province=province)

    def _country(self):
        """Returns country-level data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (object): country names
                    - Province (object): NAs
                    - City (object): NAs
                    - the other available columns
        """
        raise NotImplementedError

    def _province(self, country):
        """Returns province-level data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (object): country names
                    - Province (object): province/state/prefecture names
                    - City (object): NAs
                    - the other available columns
        """
        raise NotImplementedError

    def _city(self, country, province):
        """Returns city-level data.

        Args:
            country (str): country name
            province (str): province/state/prefecture name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (object): country names
                    - Province (object): province/state/prefecture names
                    - City (object): city names
                    - the other available columns
        """
        raise NotImplementedError

    def _provide(self, url, suffix, columns, date, date_format):
        """Provide the latest data and rename with class variable .COL_DICT.

        Args:
            url (str): URL of the data
            suffix (str): suffix of the file title
            columns (list[str]): columns to use
            date (str or None): column name of date
            date_format (str): format of date column, like %Y-%m-%d

        Returns:
            pandas.DataFrame

        Note:
            File will be downloaded to '/{self._directory}/{title}{suffix}.csv'.
        """
        filename = self._filer.csv(title=f"{self.TITLE}{suffix}")["path_or_buf"]
        df = self._provider.latest(filename=filename, url=url, columns=columns, date=date, date_format=date_format)
        return df.rename(columns=self.COL_DICT)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
from multiprocessing import cpu_count, Pool
import pandas as pd
from covsirphy.util.error import deprecate
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy._deprecated.population import PopulationData
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.phase_unit import PhaseUnit


class MPEstimator(Term):
    """
    Deprecated.
    Perform multiprocessing of Phaseunit.estimate()

    Args:
        model (covsirphy.ModelBase or None): ODE model
        jhu_data (covsirphy.JHUData): object of records
        population_data (covsirphy.PopulationData): PopulationData object
        record_df (pandas.DataFrame)
            Index
                reset index
            Columns
                - Date (pd.Timestamp): Observation date
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases
                - any other columns will be ignored
        tau (int or None): tau value [min], a divisor of 1440
        kwargs: keyword arguments of model parameters

    Note:
        When @record_df is None, @jhu_data and @population_data must be specified.
    """

    @deprecate("MPEstimator", new="ODEHandler", version="2.19.1-zeta-fu1")
    def __init__(self, model, jhu_data=None, population_data=None,
                 record_df=None, tau=None, **kwargs):
        # Records
        if jhu_data is not None and population_data is not None:
            self.jhu_data = self._ensure_instance(
                jhu_data, JHUData, name="jhu_data")
            # Population
            self.population_data = self._ensure_instance(
                population_data, PopulationData, name="population_data")
            self.from_dataset = True
        else:
            self._ensure_dataframe(record_df, name="record_df", columns=self.NLOC_COLUMNS)
            self.record_df = record_df.copy()
            self.from_dataset = False
        # Arguments
        self.model = Validator(model, "model").subclass(ModelBase)
        self._tau = Validator(tau, "tau").tau(default=None)
        self.param_dict = {
            k: v for (k, v) in kwargs.items() if k in model.PARAMETERS}
        self._units = []

    @staticmethod
    def _ensure_instance(target, class_obj, name="target"):
        """
        Ensure the target is an instance of the class object.

        Args:
            target (instance): target to ensure
            parent (class): class object
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        s = f"@{name} must be an instance of {class_obj}, but {type(target)} was applied."
        if not isinstance(target, class_obj):
            raise TypeError(s)
        return target

    @staticmethod
    def _ensure_dataframe(target, name="df", time_index=False, columns=None, empty_ok=True):
        """
        Ensure the dataframe has the columns.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            name (str): argument name of the dataframe
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            empty_ok (bool): whether give permission to empty dataframe or not

        Returns:
            pandas.DataFrame:
                Index
                    as-is
                Columns:
                    columns specified with @columns or all columns of @target (when @columns is None)
        """
        if not isinstance(target, pd.DataFrame):
            raise TypeError(f"@{name} must be a instance of (pandas.DataFrame).")
        df = target.copy()
        if time_index and (not isinstance(df.index, pd.DatetimeIndex)):
            raise TypeError(f"Index of @{name} must be <pd.DatetimeIndex>.")
        if not empty_ok and target.empty:
            raise ValueError(f"@{name} must not be a empty dataframe.")
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            cols_str = ", ".join(col for col in columns if col not in df.columns)
            included = ", ".join(df.columns.tolist())
            s1 = f"Expected columns were not included in {name} with {included}."
            raise KeyError(f"{s1} {cols_str} must be included.")
        return df.loc[:, columns]

    @property
    def tau(self):
        """
        int or None: tau value [min]
        """
        return self._tau

    def add(self, units):
        """
        Register PhaseUnits.

        Args:
            units (list[covsirphy.PhaseUnit]): list of phases

        Returns:
            covsirphy.MPEstimator: self
        """
        if not isinstance(units, list):
            raise TypeError("@units must be a list of PhaseUnits.")
        type_ok = all(isinstance(unit, PhaseUnit) for unit in units)
        if not type_ok:
            raise TypeError("@units must be a list of PhaseUnits.")
        units = [
            unit.set_ode(model=self.model, **self.param_dict) for unit in units
        ]
        self._units.extend(units)
        return self

    def _run(self, unit, tau, auto_complement=False, **kwargs):
        """
        Run estimation for one phase.

        Args:
            unit (covsirphy.PhaseUnit): unit of one phase
            tau (int or None): tau value [min], a divisor of 1440
            auto_complement (bool): if True and necessary, the number of cases will be complemented
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()
        """
        # Set tau
        unit.set_ode(tau=tau)
        # Parameter estimation
        if self.from_dataset:
            id_dict = unit.id_dict.copy()
            try:
                country = id_dict["country"]
            except KeyError:
                raise KeyError("PhaseUnit.id_dict['country'] must have country name.") from None
            province = id_dict["province"] if "province" in id_dict else None
            population = self.population_data.value(
                country=country, province=province)
            record_df, _ = self.jhu_data.records(
                country=country, province=province, population=population,
                auto_complement=auto_complement)
        else:
            record_df = self.record_df.copy()
        unit.estimate(record_df=record_df, **kwargs)
        # Show the number of trials and runtime
        unit_dict = unit.to_dict()
        trials, runtime = unit_dict[self.TRIALS], unit_dict[self.RUNTIME]
        print(f"\t{unit}: finished {trials:>4} trials in {runtime}")
        return unit

    def run(self, n_jobs=-1, **kwargs):
        """
        Run estimation.

        Args:
            n_jobs (int): the number of parallel jobs or -1 (CPU count)
            kwargs: keyword arguments of model parameters and covsirphy.Estimator.run()

        Returns:
            list[covsirphy.PhaseUnit]
        """
        units = self._units[:]
        results = []
        # The number of parallel jobs
        n_jobs = cpu_count() if n_jobs == -1 else n_jobs
        # Start optimization
        print(f"\n<{self.model.NAME} model: parameter estimation>")
        print(f"Running optimization with {n_jobs} CPUs...")
        stopwatch = StopWatch()
        # Estimation of the last phase will be done to determine tau value
        if self._tau is None:
            unit_sel, units = units[-1], units[:-1]
            unit_est = self._run(unit=unit_sel, tau=None, **kwargs)
            self._tau = unit_est.tau
            results = [unit_est]
        # Estimation of each phase
        est_f = functools.partial(self._run, tau=self._tau, **kwargs)
        if n_jobs == 1:
            results = [est_f(unit) for unit in units]
        else:
            with Pool(n_jobs) as p:
                units_est = p.map(est_f, units)
            results.extend(units_est)
        # Completion
        stopwatch.stop()
        print(f"Completed optimization. Total: {stopwatch.stop_show()}")
        return results



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['covsirphy',
 'covsirphy._deprecated',
 'covsirphy.cleaning',
 'covsirphy.downloading',
 'covsirphy.dynamics',
 'covsirphy.engineering',
 'covsirphy.gis',
 'covsirphy.science',
 'covsirphy.util',
 'covsirphy.visualization']

package_data = \
{'': ['*']}

install_requires = \
['AutoTS>=0.4,<0.6',
 'Unidecode>=1.2.0,<2.0.0',
 'better-exceptions>=0.3.3,<0.4.0',
 'country-converter>=0.7.1,<0.8.0',
 'geopandas>=0.9,<0.11',
 'japanmap>=0.0.21,<0.0.25',
 'lightgbm>=3.2.1,<4.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.21.0,<2.0.0',
 'optuna>=3.0.0,<4.0.0',
 'p-tqdm>=1.3.3,<2.0.0',
 'pandas==1.3.5',
 'pca>=1.8.2,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'ruptures>=1.1.1,<2.0.0',
 'scikit-learn>=0.24,<1.1',
 'scipy>=1.7.3,<2.0.0',
 'seaborn>=0.11.1,<0.13.0',
 'tabulate>=0.8.7,<0.9.0',
 'wbdata>=0.3.0,<0.4.0']

extras_require = \
{':python_version < "3.10"': ['datatable>=1.0.0,<2.0.0']}

setup_kwargs = {
    'name': 'covsirphy',
    'version': '2.26.2',
    'description': 'COVID-19 data analysis with phase-dependent SIR-derived ODE models',
    'long_description': '\n<img src="https://raw.githubusercontent.com/lisphilar/covid19-sir/master/docs/logo/covsirphy_headline.png" width="390" alt="CovsirPhy: COVID-19 analysis with phase-dependent SIRs">\n\n[![PyPI version](https://badge.fury.io/py/covsirphy.svg)](https://badge.fury.io/py/covsirphy)\n[![Downloads](https://pepy.tech/badge/covsirphy)](https://pepy.tech/project/covsirphy)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/covsirphy)](https://badge.fury.io/py/covsirphy)\n[![Quality Check](https://github.com/lisphilar/covid19-sir/actions/workflows/test.yml/badge.svg)](https://github.com/lisphilar/covid19-sir/actions/workflows/test.yml)\n[![GitHub license](https://img.shields.io/github/license/lisphilar/covid19-sir)](https://github.com/lisphilar/covid19-sir/blob/master/LICENSE)\n[![Test Coverage](https://codecov.io/gh/lisphilar/covid19-sir/branch/master/graph/badge.svg?token=9Z8Z1UHY3I)](https://codecov.io/gh/lisphilar/covid19-sir)\n\n# CovsirPhy introduction\n\n[<strong>Documentation</strong>](https://lisphilar.github.io/covid19-sir/index.html)\n| [<strong>Installation</strong>](https://lisphilar.github.io/covid19-sir/INSTALLATION.html)\n| [<strong>Tutorial</strong>](<https://lisphilar.github.io/covid19-sir/01_data_preparation.html>)\n| [<strong>API reference</strong>](https://lisphilar.github.io/covid19-sir/covsirphy.html)\n| [<strong>GitHub</strong>](https://github.com/lisphilar/covid19-sir)\n| [<strong>Qiita (Japanese)</strong>](https://qiita.com/tags/covsirphy)\n\n<strong>CovsirPhy is a Python library for COVID-19 (Coronavirus disease 2019) data analysis with phase-dependent SIR-derived ODE models. We can download datasets and analyze them easily. Scenario analysis with CovsirPhy enables us to make data-informed decisions. </strong>\n\n## Inspiration\n\n* Monitor the spread of COVID-19 with SIR-derived ODE models\n* Predict the number of cases in each country/province\n* Find the relationship of reproductive number and measures taken by each country\n\n<strong>If you have ideas or need new functionalities, please join this project.\nAny suggestions with [Github Issues](https://github.com/lisphilar/covid19-sir/issues/new/choose) and [Twitter: @lisphilar](https://twitter.com/lisphilar) are always welcomed. Questions are also great. Please refer to [Guideline of contribution](https://lisphilar.github.io/covid19-sir/CONTRIBUTING.html).</strong>\n\n## Installation\n\nThe latest stable version of CovsirPhy is available at [PyPI (The Python Package Index): covsirphy](https://pypi.org/project/covsirphy/) and supports Python 3.7.12 or newer versions. Details are explained in [Documentation: Installation](https://lisphilar.github.io/covid19-sir/INSTALLATION.html).\n\n```Bash\npip install --upgrade covsirphy\n```\n\nPython 3.10 users may need to install `datatable` with the following before `covsirphy` installation per [Issue #1159](https://github.com/lisphilar/covid19-sir/issues/1159).\n\n```Bash\npip install git+https://github.com/h2oai/datatable.git\n```\n\n## Demo\n\nQuickest tour of CovsirPhy is here. The following codes analyze the records in Japan.\n\n```Python\nimport covsirphy as cs\n# Data preparation,time-series segmentation, parameter estimation with SIR-F model\nsnr = cs.ODEScenario.auto_build(geo="Japan", model=cs.SIRFModel)\n# Check actual records\nsnr.simulate(name=None);\n# Show the result of time-series segmentation\nsnr.to_dynamics(name="Baseline").detect();\n# Perform simulation with estimated ODE parameter values\nsnr.simulate(name="Baseline");\n# Predict ODE parameter values (30 days from the last date of actual records)\nsnr.build_with_template(name="Predicted", template="Baseline");\nsnr.predict(days=30, name="Predicted");\n# Perform simulation with estimated and predicted ODE parameter values\nsnr.simulate(name="Predicted");\n# Add a future phase to the baseline (ODE parameters will not be changed)\nsnr.append();\n# Show created phases and ODE parameter values\nsnr.summary()\n# Compare reproduction number of scenarios (predicted/baseline)\nsnr.compare_param("Rt");\n# Compare simulated number of cases\nsnr.compare_cases("Confirmed");\n# Describe representative values\nsnr.describe()\n```\n\nOutput of `snr.simulate(name="Predicted");`\n\n<img src="https://raw.githubusercontent.com/lisphilar/covid19-sir/master/example/output/demo_jpn/04_predicted.png" width="600">\n\n## Tutorial\n\nTutorials of functionalities are included in the [CovsirPhy documentation](https://lisphilar.github.io/covid19-sir/index.html).\n\n* [Data preparation](https://lisphilar.github.io/covid19-sir/01_data_preparation.html)\n* [Data Engineering](https://lisphilar.github.io/covid19-sir/02_data_engineering.html)\n* [SIR-derived ODE models](https://lisphilar.github.io/covid19-sir/03_ode.html)\n* [Phase-dependent SIR models](https://lisphilar.github.io/covid19-sir/04_phase_dependent.html)\n* [Scenario analysis](https://lisphilar.github.io/covid19-sir/05_scenario_analysis.html)\n* [ODE parameter prediction](https://lisphilar.github.io/covid19-sir/06_prediction.html)\n\n## Release notes\n\nRelease notes are [here](https://github.com/lisphilar/covid19-sir/releases). Titles & links of issues are listed with acknowledgement.\n\nWe can see the release plan for the next stable version in [milestone page of the GitHub repository](https://github.com/lisphilar/covid19-sir/milestones). If you find a highly urgent matter, please let us know via [issue page](https://github.com/lisphilar/covid19-sir/issues).\n\n## Developers\n\nCovsirPhy library is developed by a community of volunteers. Please see the full list [here](https://github.com/lisphilar/covid19-sir/graphs/contributors).\n\nThis project started in Kaggle platform. Hirokazu Takaya ([@lisphilar](<https://www.kaggle.com/lisphilar>)) published [Kaggle Notebook: COVID-19 data with SIR model](https://www.kaggle.com/lisphilar/covid-19-data-with-sir-model) on 12Feb2020 and developed it, discussing with Kaggle community. On 07May2020, "covid19-sir" repository was created. On 10May2020, `covsirphy` version 1.0.0 was published in GitHub. First release in PyPI (version 2.3.0) was on 28Jun2020.\n\n## Support\n\nPlease support this project as a developer (or a backer).\n[![Become a backer](https://opencollective.com/covsirphy/tiers/backer.svg?avatarHeight=36&width=600)](https://opencollective.com/covsirphy)\n\n## License: Apache License 2.0\n\nPlease refer to [LICENSE](https://github.com/lisphilar/covid19-sir/blob/master/LICENSE) file.\n\n## Citation\n\nPlease cite this library as follows with version number (`import covsirphy as cs; cs.__version__`).\n\n**Hirokazu Takaya and CovsirPhy Development Team (2020-2022), CovsirPhy version [version number]: Python library for COVID-19 analysis with phase-dependent SIR-derived ODE models, [https://github.com/lisphilar/covid19-sir](https://github.com/lisphilar/covid19-sir)**\n\nThis is the output of `covsirphy.__citation__`.\n\n```Python\nimport covsirphy as cs\ncs.__citation__\n```\n\n**We have no original papers the author and contributors wrote, but note that some scientific approaches, including SIR-F model, S-R change point analysis, phase-dependent approach to SIR-derived models, were developed in this project.**\n',
    'author': 'Hirokazu Takaya',
    'author_email': 'lisphilar@outlook.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lisphilar/covid19-sir',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.12,<3.11',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from copy import deepcopy
from datetime import timedelta
import re
import pandas as pd
from covsirphy.util.error import ScenarioNotFoundError, SubsetNotFoundError, UnExpectedTypeError, UnExpectedValueRangeError
from covsirphy.util.error import UnExpectedValueError
from covsirphy.util.validator import Validator
from covsirphy.util.alias import Alias
from covsirphy.util.term import Term
from covsirphy.gis.gis import GIS
from covsirphy.visualization.line_plot import line_plot
from covsirphy.engineering.engineer import DataEngineer
from covsirphy.dynamics.ode import ODEModel
from covsirphy.dynamics.dynamics import Dynamics
from covsirphy.science.ml import MLEngineer


class ODEScenario(Term):
    """Perform scenario analysis, changing ODE parameters.

    Args:
        data (pandas.DataFrame): actual data of the number of cases
            Index
                Date (pandas.Timestamp): observation dates
            Columns
                Population (int): total population
                Confirmed (int): the number of confirmed cases
                Recovered (int): the number of recovered cases, must be over 0
                Fatal (int): the number of fatal cases
                Susceptible (int): the number of susceptible cases, will be ignored because overwritten
                Infected (int): the number of currently infected cases, will be ignored because overwritten
                the other columns will be ignored
        location_name (str): name to identify the location to show in figure titles
        complement (bool): perform data complement with covsirphy.DataEngineer().subset(complement=True) or not

    Note:
        Data cleaning will be performed with covsirphy.DataEngineer().clean() automatically.
    """
    _PARAM = "param"

    def __init__(self, data, location_name, complement=True):
        self._location_name = str(location_name)
        # Actual records: Date index, S/I/F/R
        df = Validator(data, "data", accept_none=False).dataframe(
            time_index=True, columns=[self.N, self.C, self.F, self.R])
        df.index.name = self.DATE
        df["location"] = self._location_name
        engineer = DataEngineer(layers=["location"])
        engineer.register(data=df.reset_index())
        engineer.clean()
        engineer.transform()
        self._actual_df, *_ = engineer.subset(
            geo=self._location_name, variables=[self.S, self.CI, self.R, self.F], complement=complement)
        self._first, self._last = self._actual_df.index.min(), self._actual_df.index.max()
        # {scenario_name: {"ODE": ODEModel, "tau": int, "param": pd.DataFrame(index: Date, columns: ODE parameters)}}
        self._snr_alias = Alias(target_class=dict)
        # Aliases of variable names
        self._variable_alias = Alias.for_variables()

    def build_with_dynamics(self, name, dynamics):
        """Build a scenario with covsirphy.Dynamics() instance.

        Args:
            name (str): scenario name
            dynamics (covsirphy.Dynamics): covsirphy.Dynamics() instance which has ODE model, tau value and ODE parameter values

        Return:
            covsirphy.ODEScenario: self
        """
        dyn = Validator(dynamics, "dynamics").instance(Dynamics)
        snl_dict = {self.ODE: dyn.model, self.TAU: dyn.tau, self._PARAM: dyn.track().loc[:, dyn.model._PARAMETERS]}
        self._snr_alias.update(name=name, target=snl_dict)
        return self

    def build_with_model(self, name, model, date_range=None, tau=None):
        """Build a scenario with covsirphy.Dynamics() instance created with the actual data automatically.

        Args:
            name (str): scenario name
            model (covsirphy.ODEModel): definition of ODE model
            date_range (tuple of (str, str) or None): start date and end date of dynamics to analyze
            tau (int or None): tau value [min] or None (set later with data)

        Return:
            covsirphy.ODEScenario: self
        """
        Validator(model, "model").subclass(ODEModel)
        start_date, end_date = Validator(date_range, "date_range").sequence(default=(None, None), length=2)
        start = Validator(start_date, name="the first value of @date_range").date(default=self._first)
        end = Validator(
            end_date, name="the second date of @date_range").date(value_range=(start, None), default=self._last)
        dyn = Dynamics(model=model, date_range=(start, end), tau=tau)
        dyn.register(data=self._actual_df.loc[:, self._SIRF])
        dyn.segment(points=None, overwrite=True, display=False)
        dyn.estimate()
        return self.build_with_dynamics(name=name, dynamics=dyn)

    def build_with_template(self, name, template):
        """Build a scenario with a template scenario.

        Args:
            name (str): new scenario name
            template (str): template name

        Raises:
            SubsetNotFoundError: scenario with the name is un-registered

        Return:
            covsirphy.ODEScenario: self
        """
        temp_snl_dict = self._snr_alias.find(name=template)
        if temp_snl_dict is None:
            raise ScenarioNotFoundError(name=template)
        self._snr_alias.update(name=name, target=deepcopy(temp_snl_dict))
        return self

    @classmethod
    def auto_build(cls, geo, model, complement=True):
        """Prepare cleaned and subset data from recommended dataset, create instance, build baseline scenario.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): country, province, city
            model (covsirphy.ODEModel): definition of ODE model
            complement (bool): whether perform complement or not

        Raises:
            SubsetNotFoundError: actual data of the location was not included in the recommended dataset

        Return:
            covsirphy.ODEScenario: created instance

        Note:
            `geo=None` means total values of all countries.

        Note:
            `geo="Japan"` and `geo=("Japan",)` means country level data of Japan, as an example.

        Note:
            `geo=("Japan", "Tokyo")` means prefecture (province) level data of Tokyo/Japan, as an example.

        Note:
            `geo=("USA", "Alabama", "Baldwin")` means country level data of Baldwin/Alabama/USA, as an example.

        Note:
            Complemented (if @complement is True) data with Recovered > 0 will be analyzed.
        """
        Validator(geo, "geo", accept_none=True).instance(expected=(str, tuple, list))
        Validator(model, "model", accept_none=False).subclass(ODEModel)
        # Prepare data
        engineer = DataEngineer()
        engineer.download(
            country=geo[0] if isinstance(geo, (tuple, list)) and len(geo) > 1 else None,
            province=geo[1] if isinstance(geo, (tuple, list)) and len(geo) > 2 else None,
            databases=["japan", "covid19dh"])
        engineer.clean()
        try:
            subset_df, *_ = engineer.subset(geo=geo, complement=complement)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                geo=geo, details="Please create covsirphy.DataEngineer() instance to prepare data") from None
        # Create instance
        snl = cls(data=subset_df.loc[subset_df[cls.R] > 0], location_name=GIS.area_name(geo=geo), complement=False)
        # Build baseline scenario
        snl.build_with_model(name="Baseline", model=model)
        return snl

    def delete(self, pattern, exact=False):
        """Delete scenario(s).

        Args:
            pattern (str): scenario name or pattern to search
            exact (bool): if False, use regular expressions

        Return:
            covsirphy.ODEScenario: self
        """
        if exact:
            self._snr_alias.delete(name=pattern)
            return self
        p = re.compile(pattern)
        names = [name for name in self._snr_alias.all().keys() if p.search(name)]
        for name in names:
            self._snr_alias.delete(name=name)
        return self

    def rename(self, old, new):
        """Rename the given scenario names with a new one.

        Args:
            old (str): old name
            new (str): new name

        Returns:
            covsirphy.Scenario: self
        """
        self.build_with_template(name=new, template=old)
        self.delete(pattern=old, exact=True)
        return self

    def to_dynamics(self, name):
        """Create covsirphy.Dynamics instance of the scenario.

        Args:
            name (str): scenario name

        Raises:
            SubsetNotFoundError: scenario with the name is un-registered

        Returns:
            covsirphy.Dynamics: instance which has ODE model, tau value and ODE parameter values
        """
        temp_snl_dict = self._snr_alias.find(name=name)
        if temp_snl_dict is None:
            raise ScenarioNotFoundError(name=name)
        model, tau, param_df = [temp_snl_dict[k] for k in [self.ODE, self.TAU, self._PARAM]]
        df = self._actual_df.join(param_df, how="right")
        return Dynamics.from_data(model=model, data=df, tau=tau, name=name)

    def summary(self):
        """Summarize phase information of all scenarios.

        Returns:
            pandas.DataFrame:
                Index
                    Scenario (str): scenario names
                    Phase (str): phase names, 0th, 1st,...
                Columns
                    Start (pandas.Timestamp): start date of the phase
                    End (pandas.Timestamp): end date of the phase
                    ODE (str): ODE model name
                    Rt (float): phase-dependent reproduction number (if parameters are available)
                    (float): parameter values, including rho (if available)
                    tau (int): tau value [min]
                    (int or float): dimensional parameters, including 1/beta [days] (if tau and parameters are available)
        """
        dataframes = []
        for name, snl_dict in self._snr_alias.all().items():
            dyn = self.to_dynamics(name=name)
            df = dyn.summary().reset_index()
            df[self.SERIES] = name
            df[self.ODE] = snl_dict[self.ODE].name()
            df[self.TAU] = snl_dict[self.TAU]
            dataframes.append(df)
        return pd.concat(dataframes, axis=0).set_index([self.SERIES, self.PHASE]).convert_dtypes()

    def track(self):
        """Track reproduction number, parameter value and dimensional parameter values.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    Scenario (str): scenario names
                    Phase (str): phase names
                    Date (pandas.Timestamp): dates
                    Rt (float): phase-dependent reproduction number (if parameters are available)
                    (float): parameter values, including rho (if available)
                    (int or float): dimensional parameters, including 1/beta [days] (if tau and parameters are available)
        """
        df = self.summary().reset_index(drop=False)
        df[self.DATE] = df[[self.START, self.END]].apply(
            lambda x: pd.date_range(start=x[0], end=x[1], freq="D"), axis=1)
        return df.explode(self.DATE).set_index(self.DATE).drop([self.START, self.END], axis=1)

    def describe(self):
        """Describe representative values.

        Returns:
            pandas.DataFrame:
                Index
                    str: scenario name
                Columns
                    - max(Infected) (numpy.int64): max value of Infected
                    - argmax(Infected) (pandas.Timestamp): the date when Infected shows max value
                    - Confirmed({date}) (numpy.int64): Confirmed on the last date
                    - Infected({date} (numpy.int64)): Infected on the last date
                    - Fatal({date}) (numpy.int64): Fatal on the last date
        """
        _dict = {}
        for name in self._snr_alias.all().keys():
            dyn = self.to_dynamics(name=name)
            sim_df = dyn.simulate(model_specific=False)
            sim_df[self.SERIES] = name
            engineer = DataEngineer(layers=[self.SERIES])
            engineer.register(data=sim_df.reset_index())
            engineer.inverse_transform()
            sim_df = engineer.all().set_index(self.DATE)
            last_date = sim_df.index[-1]
            last_date_str = last_date.strftime(self.DATE_FORMAT)
            _dict[name] = {
                f"max({self.CI})": sim_df[self.CI].max(),
                f"argmax({self.CI})": sim_df[self.CI].idxmax(),
                f"{self.C} on {last_date_str}": sim_df.loc[last_date, self.C],
                f"{self.CI} on {last_date_str}": sim_df.loc[last_date, self.CI],
                f"{self.F} on {last_date_str}": sim_df.loc[last_date, self.F],
            }
        return pd.DataFrame.from_dict(_dict, orient="index")

    def simulate(self, name=None, variables=None, display=True, **kwargs):
        """Perform simulation with phase-dependent ODE model.

        Args:
            name (str or None): scenario name registered or None (actual data)
            variables (list of [str] or None): variables/alias to return or None (["Confirmed", "Fatal", "Recovered"])
            display (bool): whether display figure of the result or not
            **kwargs: keyword arguments of covsirphy.line_plot() except for @df

        Returns:
            pandas.DataFrame or pandas.Series:
                Index
                    Date (pd.Timestamp): dates
                Columns
                    Population (int): total population (if selected with @variables)
                    Confirmed (int): the number of confirmed cases (if selected with @variables)
                    Recovered (int): the number of recovered cases (if selected with @variables)
                    Fatal (int): the number of fatal cases (if selected with @variables)
                    Susceptible (int): the number of susceptible cases (if selected with @variables)
                    Infected (int): the number of currently infected cases (if selected with @variables)
        """
        if name is None:
            sirf_df = self._actual_df.copy()
            title = f"{self._location_name}: actual number of cases over time"
            v = None
        else:
            dyn = self.to_dynamics(name=name)
            sirf_df = dyn.simulate(model_specific=False)
            title = f"{self._location_name} ({name} scenario): simulated number of cases over time"
            v = dyn.start_dates()[1:]
        sirf_df[self.SERIES] = name or self.ACTUAL
        engineer = DataEngineer(layers=[self.SERIES])
        engineer.register(data=sirf_df.reset_index())
        engineer.inverse_transform()
        v_converted = self._variable_alias.find(name=variables, default=[self.C, self.F, self.R])
        df = engineer.all().set_index(self.DATE).loc[self._first:self._last, v_converted]
        # Show figure
        if display:
            plot_kwargs = {"title": title, "y_integer": True, "v": v, "ylabel": "the number of cases", **kwargs}
            line_plot(df=df, **plot_kwargs)
        return df

    def compare_cases(self, variable, date_range=None, ref=None, display=True, **kwargs):
        """Compare the number of cases of scenarios.

        Args:
            variable (str): variable name or alias
            date_range (tuple of (str, str)): start date and end date to analyze
            ref (str or None): name of reference scenario to specify phases and dates or None (the first scenario)
            display (bool): whether display figure of the result or not
            **kwargs: keyword arguments of covsirphy.line_plot() except for @df

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): dates
                Columns
                    Actual (numpy.int64): actual records
                    {scenario name} (numpy.int64): values of the scenario
        """
        v_converted = Validator(
            self._variable_alias.find(name=variable, default=[variable]), "variable", accept_none=False).sequence(
            length=1, candidates=[self.C, self.N, *self._actual_df.columns.tolist()])[0]
        df = pd.DataFrame(index=self._actual_df.index.tolist())
        if v_converted == self.C:
            df[self.ACTUAL] = self._actual_df[[self.CI, self.F, self.R]].sum(axis=1)
        elif v_converted == self.N:
            df[self.ACTUAL] = self._actual_df[[self.CI, self.F, self.R, self.S]].sum(axis=1)
        else:
            df[self.ACTUAL] = self._actual_df.loc[:, v_converted]
        dataframes = [
            self.simulate(name=name, display=False)[v_converted].rename(name) for name in self._snr_alias.all().keys()]
        df = pd.concat([df, *dataframes], axis=1)
        start_date, end_date = Validator(date_range, "date_range").sequence(default=(None, None), length=2)
        start = Validator(start_date, name="the first value of @date_range").date(default=df.index.min())
        end = Validator(
            end_date, name="the second date of @date_range").date(value_range=(start, None), default=df.index.max())
        df = df.loc[start: end]
        if display:
            ylabel = f"the number of {v_converted.lower()} cases"
            title = f"{self._location_name}: {ylabel} overt time"
            v = self.to_dynamics(name=ref or list(self._snr_alias.all().keys())[0]).start_dates()[1:]
            plot_kwargs = {"title": title, "y_integer": True, "v": v, "ylabel": ylabel, **kwargs}
            line_plot(df=df, **plot_kwargs)
        return df.convert_dtypes()

    def compare_param(self, param, date_range=None, ref=None, display=True, **kwargs):
        """Compare the number of cases of scenarios.

        Args:
            param (str): one of ODE parameters, "Rt", dimensional parameters
            date_range (tuple of (str, str)): start date and end date to analyze
            ref (str or None): name of reference scenario to specify phases and dates or None (the first scenario)
            display (bool): whether display figure of the result or not
            **kwargs: keyword arguments of covsirphy.line_plot() except for @df

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp)
                Columns
                    {scenario name} (str): values of the scenario
        """
        df = self.track()
        Validator([param], "param", accept_none=False).sequence(candidates=df.columns)
        df[param] = df[param].astype("float64")
        df = df.pivot_table(values=param, index=self.DATE, columns=self.SERIES)
        start_date, end_date = Validator(date_range, "date_range").sequence(default=(None, None), length=2)
        start = Validator(start_date, name="the first value of @date_range").date(default=df.index.min())
        end = Validator(
            end_date, name="the second date of @date_range").date(value_range=(start, None), default=df.index.max())
        df = df.loc[start: end]
        if display:
            ylabel, h = (self.RT_FULL, 1.0) if param == self.RT else (param, None)
            title = f"{self._location_name}: {ylabel} overt time"
            v = self.to_dynamics(name=ref or list(self._snr_alias.all().keys())[0]).start_dates()[1:]
            plot_kwargs = {"title": title, "math_scale": False, "v": v, "ylabel": ylabel, "h": h, **kwargs}
            line_plot(df=df, **plot_kwargs)
        return df.convert_dtypes()

    def _append(self, name, end, **kwargs):
        """Append a new phase, specifying ODE parameter values.

        Args:
            name (str): scenario name
            end (pandas.Timestamp or int): end date or the number days of new phase
            **kwargs: keyword arguments of ODE parameter values (default: values of the last phase)

        Raises:
            SubsetNotFoundError: scenario with the name is un-registered
            UnExpectedValueRangeError: end_date - (the last date of the registered phases) < 3 and parameters were changed
        """
        try:
            snr_dict = self._snr_alias.find(name=name).copy()
        except AttributeError:
            raise ScenarioNotFoundError(name=name) from None
        param_df = snr_dict[self._PARAM].copy()
        last_param_dict = param_df.iloc[-1].to_dict()
        start_date = param_df.index[-1] + timedelta(days=1)
        try:
            delta = timedelta(days=Validator(end, "end", accept_none=False).int(value_range=(0, None)))
            end_date = param_df.index[-1] + delta
        except UnExpectedTypeError:
            end_date = Validator(end, "end", accept_none=False).date(value_range=(param_df.index[-1], None))
        new_param_dict = Validator(kwargs, "keyword arguments").dict(
            default=last_param_dict, required_keys=list(last_param_dict.keys()))
        new_df = pd.DataFrame(new_param_dict, index=pd.date_range(start=start_date, end=end_date, freq="D"))
        snr_dict[self._PARAM] = pd.concat([param_df, new_df], axis=0)
        try:
            Dynamics.from_data(
                model=snr_dict[self.ODE],
                data=self._actual_df.join(snr_dict[self._PARAM], how="right"),
                tau=snr_dict[self.TAU])
        except UnExpectedValueError:
            if isinstance(end, int):
                target, _min = end, 3
            else:
                target = end.strftime(self.DATE_FORMAT)
                _min = (param_df.index[-1] + timedelta(days=3)).strftime(self.DATE_FORMAT)
            raise UnExpectedValueRangeError(
                name="end", target=target, value_range=(_min, None)) from None
        self._snr_alias.update(name=name, target=snr_dict.copy())
        self._last = max((self._last, end_date))

    def append(self, name=None, end=None, **kwargs):
        """Append a new phase, specifying ODE parameter values.

        Args:
            name (str or list[str] None): scenario name(s) or None (all scenarios)
            end (pandas.Timestamp or int or None): end date or the number days of new phase or None (the max date of all scenarios and actual data)
            **kwargs: keyword arguments of ODE parameter values (default: values of the last phase)

        Raises:
            SubsetNotFoundError: scenario with the name is un-registered
            UnExpectedValueRangeError: end_date - (the last date of the registered phases) < 3 and parameters were changed

        Return:
            covsirphy.ODEScenario: self
        """
        if end is None:
            last_end = self._last
            for snl_dict in self._snr_alias.all().values():
                end_date = snl_dict[self._PARAM].index[-1]
                last_end = max(last_end, end_date)
        names = [name] if isinstance(name, str) else Validator(name, "name").sequence(
            default=list(self._snr_alias.all().keys()))
        for _name in names:
            try:
                self._append(name=_name, end=end or last_end, **kwargs)
            except UnExpectedValueRangeError as e:
                raise e from None
        return self

    def predict(self, days, name, seed=0, verbose=1, X=None, **kwargs):
        """Create scenarios and append a phase, performing prediction ODE parameter prediction for given days.

        Args:
            days (int): days to predict
            name (str): scenario name
            X (pandas.DataFrame or None): information for regression or None (no information)
                Index
                    pandas.Timestamp: Observation date
                Columns
                    observed and the target variables (int or float)
            seed (int or None): random seed
            verbose (int): verbosity
            **kwargs: keyword arguments of autots.AutoTS() except for verbose, forecast_length (always the same as @days)

        Return:
            covsirphy.ODEScenario: self

        Note:
            AutoTS package is developed at https://github.com/winedarksea/AutoTS

        Note:
            Phases are determined with rounded reproduction number (one decimal place).
        """
        model = self._snr_alias.find(name=name)[self.ODE]
        Y = self.to_dynamics(name=name).track().loc[:, model._PARAMETERS]
        # Parameter prediction
        eng = MLEngineer(seed=seed, verbose=verbose)
        param_df = eng.forecast(Y=Y, days=days, X=X, **kwargs).reset_index()
        # Create phases with Rt values
        param_df[self.RT] = param_df[model._PARAMETERS].apply(
            lambda x: model.from_data(data=self._actual_df.reset_index(), param_dict=x.to_dict(), tau=1440).r0(), axis=1)
        param_df["Phase"] = (param_df[self.RT] != param_df[self.RT].shift()).cumsum()
        phase_df = param_df.groupby("Phase").last().drop(self.RT, axis=1).rename(columns={self.DATE: "end"})
        phase_df = phase_df.loc[
            (phase_df.index[-1]) | (phase_df["end"] < phase_df["end"].shift(periods=-1) - timedelta(days=3))]
        for phase_dict in phase_df.to_dict(orient="records"):
            with contextlib.suppress(UnExpectedValueRangeError):
                # Skip when the first new phase is smaller than 3 days
                self._append(name=name, **phase_dict)
        return self

    def represent(self, q, variable, date=None, included=None, excluded=None):
        """
        Return the names of representative scenarios using quantiles of the variable on on the date.

        Args:
            q (list[float] or float): quantiles
            variable (str): reference variable, Confirmed, Infected, Fatal or Recovered
            date (str or None): reference date or None (the last end date in the all scenarios)
            included (list[str] or None): included scenarios or None (all included)
            excluded (list[str] or None): excluded scenarios or None (no scenarios not excluded)

        Raises:
            ValueError: the end dates of the last phase is not aligned

        Returns:
            list[float] or float: the nearest scenarios which has the values at the given quantiles

        Note:
            Dimension of returned object corresponds to the type of @q.
        """
        quantiles = [q] if isinstance(q, float) else Validator(q, "q").sequence()
        v_converted = self._variable_alias.find(name=variable, default=[variable])
        Validator(v_converted, "variable", accept_none=False).sequence(length=1)
        # Target scenario to included
        all_set = set(self._snr_alias.all().keys())
        in_set = all_set if included is None else set(Validator(included, "included").sequence())
        ex_set = set() if excluded is None else set(Validator(excluded, "excluded").sequence())
        scenarios = list(all_set & (in_set) - ex_set)
        # Get simulation data of the variable of the target scenarios
        sim_dict = {name: self.simulate(name=name, display=False)[variable] for name in scenarios}
        sim_df = pd.DataFrame(sim_dict)
        if sim_df.isna().to_numpy().sum():
            raise ValueError(
                "The end dates of the last phases must be aligned. Scenario.adjust_end() method may fix this issue.")
        # Find representative scenario
        date_obj = Validator(date, "date").date(default=sim_df.index[-1])
        series = sim_df.loc[date_obj].squeeze()
        values = series.quantile(q=quantiles, interpolation="nearest")
        return [series[series == v].index.tolist()[0] for v in values]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.error import NAFoundError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.ode_solver import _ODESolver


class _MultiPhaseODESolver(Term):
    """
    Perform simulation with a multi-phased ODE model.

    Args:
        model (covsirphy.ModelBase): ODE model
        first (pandas.Timestamp): first date of simulation, like 14Apr2021
        tau (int): tau value [min]
    """

    def __init__(self, model, first, tau):
        self._model = Validator(model, "model").subclass(ModelBase)
        self._first = Validator(first, "first").instance(pd.Timestamp)
        self._tau = Validator(tau, "tau").tau(default=None)
        if self._tau is None:
            NAFoundError("tau", None)
        # {"0th": output of self.add()}
        self._info_dict = {}

    def _add(self, end, param_dict, y0_dict):
        """
        Add a new phase.

        Args:
            end (pandas.Timestamp): end date of the phase
            param_dict (dict[str, float]): parameter values
            y0_dict (dict[str, int] or None): initial values or None (not set)

        Returns:
            dict(str, object): setting of the phase
                - param (dict[str, float]): parameter values
                - y0 (dict[str, int]): initial values of model-specialized variables or empty dict
                - step_n (int): the number of steps

        Note:
            Internal variable "step_n" means from the first date to the next date of the end date.
        """
        Validator(end, end).instance(pd.Timestamp)
        phase = self.num2str(len(self._info_dict))
        all_step_n = self.steps(
            self._first.strftime(self.DATE_FORMAT), end.strftime(self.DATE_FORMAT), tau=self._tau)
        step_n = all_step_n - sum(phase_dict["step_n"] for phase_dict in self._info_dict.values())
        self._info_dict[phase] = {"param": param_dict, "y0": y0_dict or {}, "step_n": step_n}
        return self._info_dict[phase]

    def simulate(self, *args):
        """
        Perform simulation with the multi-phased ODE model.

        Args:
            args (dict(str, object)): list of phase settings.
                - End (pandas.Timestamp): end date of the phase
                - param_dict (dict[str, float]): parameter values
                - y0_dict (dict[str, int] or None): initial values or None (not set)

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        # Settings
        for phase_dict in args:
            self._add(phase_dict[self.END], phase_dict["param"], phase_dict["y0"])
        # Multi-phased simulation
        dataframes = []
        for (_, phase_dict) in self._info_dict.items():
            # Step numbers
            step_n = phase_dict["step_n"]
            # Initial values: registered information (with priority) or the last values
            y0_dict = dataframes[-1].iloc[-1].to_dict() if dataframes else {}
            y0_dict.update(phase_dict["y0"])
            if set(self._model.VARIABLES) - y0_dict.keys():
                un_vars = list(set(self._model.VARIABLES) - y0_dict.keys())
                s, be = ("s", "were") if len(un_vars) > 1 else ("", "was")
                raise KeyError(f"Initial value{s} of <{'>, <'.join(un_vars)}> {be} not specified.")
            # parameter values
            param_dict = phase_dict["param"].copy()
            # Solve the initial value problem with the ODE model
            solver = _ODESolver(self._model, **param_dict)
            solved_df = solver.run(step_n=step_n, **y0_dict)
            dataframes += [solved_df.iloc[1:]] if dataframes else [solved_df]
        # Combine the simulation results
        df = pd.concat(dataframes, ignore_index=True, sort=True)
        return self._model.convert_reverse(df, start=self._first, tau=self._tau)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _DataTransformer(Term):
    """Class for data transformation.

    Args:
        data (pandas.DataFrame): raw data
            Index
                reset index
            Column
                columns defined by @layers
                column defined by @date
                the other columns
        layers (list[str]): location layers of the data
        date (str): column name of observation dates of the data
    """

    def __init__(self, data, layers, date):
        self._df = data.copy()
        self._layers = Validator(layers, "layers").sequence()
        self._date = str(date)

    def all(self):
        """Return all available data.

        Returns:
            pandas.DataFrame: transformed data
        """
        return self._df

    def susceptible(self, new, **kwargs):
        """Calculate the number of susceptible cases with "Susceptible = Population - Confirmed" formula.

        Args:
            new (str): new column name
            kwargs (dict[str, str]): dictionary of columns
                - population (str): total population
                - confirmed (str): the number of confirmed cases
        """
        c_dict = Validator(kwargs, "kwargs").dict(required_keys=["population", "confirmed"], errors="raise")
        df = Validator(self._df, "raw data").dataframe(columns=list(c_dict.values()))
        df[new] = df[c_dict["population"]] - df[c_dict["confirmed"]]
        self._df = df.copy()

    def infected(self, new, **kwargs):
        """Calculate the number of infected cases with "Infected = Confirmed - Fatal - Recovered" formula.

        Args:
            new (str): new column name
            kwargs (dict[str, str]): dictionary of columns
                - confirmed (str): the number of confirmed cases
                - fatal (str): the number of fatal cases
                - recovered (str): the number of recovered cases
        """
        c_dict = Validator(kwargs, "kwargs").dict(required_keys=["confirmed", "fatal", "recovered"], errors="raise")
        df = Validator(self._df, "raw data").dataframe(columns=list(c_dict.values()))
        df[new] = df[c_dict["confirmed"]] - df[c_dict["fatal"]] - df[c_dict["recovered"]]
        self._df = df.copy()

    def diff(self, column, suffix, freq):
        """Calculate first discrete difference of element with "F(x>0) = F(x) - F(x-1), F(0) = 0".

        Args:
            column (str): column name of the cumulative numbers
            suffix (str): suffix if the column (new column name will be '{column}{suffix}')
            freq (str): offset aliases of shifting dates

        Note:
            Regarding @freq, refer to https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        """
        df = Validator(self._df, "raw data").dataframe(columns=[column])
        new_column = f"{column}{suffix}"
        series = df.set_index(self._date).groupby(self._layers).shift(freq=freq)[column]
        series.name = new_column
        df = df.merge(series, how="left", left_on=[*self._layers, self._date], right_index=True)
        df[new_column] = (df[column] - df[new_column]).fillna(0)
        self._df = df.copy()

    def add(self, columns, new, fill_value):
        """Calculate element-wise addition with pandas.DataFrame.sum(axis=1), X1 + X2 + X3 +...

        Args:
            columns (str): columns to add
            new (str): column name of addition
            fill_value (float): value to fill in NAs
        """
        Validator(columns, "columns").sequence(candidates=list(self._df.columns))
        df = self._df.copy()
        df[new] = df[columns].sum(axis=1).fillna(Validator(fill_value, "fill_value").float())
        self._df = df.copy()

    def mul(self, columns, new, fill_value):
        """Calculate element-wise multiplication with pandas.DataFrame.product(axis=1), X1 * X2 * X3 *...

        Args:
            columns (str): columns to multiply
            new (str): column name of multiplication
            fill_value (float): value to fill in NAs
        """
        Validator(columns, "columns").sequence(candidates=list(self._df.columns))
        df = self._df.copy()
        df[new] = df[columns].product(axis=1).fillna(Validator(fill_value, "fill_value").float())
        self._df = df.copy()

    def sub(self, minuend, subtrahend, new, fill_value):
        """Calculate element-wise subtraction with pandas.Series.sub(), minuend - subtrahend.

        Args:
            minuend (str): numerator column
            subtrahend (str): subtrahend column
            new (str): column name of subtraction
            fill_value (float): value to fill in NAs
        """
        v = Validator([minuend, subtrahend], "columns of numerator and subtrahend")
        v.sequence(candidates=list(self._df.columns))
        df = self._df.copy()
        df[new] = df[minuend].sub(df[subtrahend], fill_value=Validator(fill_value, "fill_value").float())
        self._df = df.copy()

    def div(self, numerator, denominator, new, fill_value):
        """Calculate element-wise floating division with pandas.Series.div(), numerator / denominator.

        Args:
            numerator (str): numerator column
            denominator (str): denominator column
            new (str): column name of floating division
            fill_value (float): value to fill in NAs

        Note:
            Positive rate could be calculated with Confirmed / Tested, `.div(numerator="Confirmed", denominator="Tested", new="Positive_rate")`
        """
        v = Validator([numerator, denominator], "columns of numerator and denominator")
        v.sequence(candidates=list(self._df.columns))
        df = self._df.copy()
        df[new] = df[numerator].div(df[denominator], fill_value=Validator(fill_value, "fill_value").float())
        self._df = df.copy()

    def assign(self, **kwargs):
        """Assign a new column with pandas.DataFrame.assign().

        Args:
            **kwargs: dict of {str: callable or pandas.Series}

        Note:
            Refer to documentation of pandas.DataFrame.assign(), https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.assign.html
        """
        self._df = self._df.assign(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import sklearn.metrics
from covsirphy.util.error import UnExpectedValueError, NAFoundError
from covsirphy.util.validator import Validator


class Evaluator(object):
    """
    Evaluate residual errors.

    Args:
        y_true (pandas.DataFrame or pandas.Series): correct target values
        y_pred (pandas.DataFrame or pandas.Series): estimated target values
        how (str): "all" (use all records) or "inner" (intersection will be used)
        on (str or list[str] or None): column names to join on or None (join on index)

    Raises:
        NAFoundError: either @y_true or @pred has NA values

    Note:
        Evaluation with metrics will be done with sklearn.metrics package.
        https://scikit-learn.org/stable/modules/model_evaluation.html
    """
    # Names
    _A = "_actual"
    _P = "_predicted"
    # Metrics: {name: (function(x1, x2), whether smaller is better or not)}
    _METRICS_DICT = {
        "ME": (sklearn.metrics.max_error, True),
        "MAE": (sklearn.metrics.mean_absolute_error, True),
        "MSE": (sklearn.metrics.mean_squared_error, True),
        "MSLE": (sklearn.metrics.mean_squared_log_error, True),
        "MAPE": (sklearn.metrics.mean_absolute_percentage_error, True),
        "RMSE": (lambda x1, x2: sklearn.metrics.mean_squared_error(x1, x2, squared=False), True),
        "RMSLE": (lambda x1, x2: np.sqrt(sklearn.metrics.mean_squared_log_error(x1, x2)), True),
        "R2": (sklearn.metrics.r2_score, False),
    }

    def __init__(self, y_true, y_pred, how="inner", on=None):
        # Check types
        for (y, name) in zip([y_true, y_pred], ["y_true", "y_pred"]):
            Validator(y, name, accept_none=False).instance(expected=(pd.DataFrame, pd.Series, list, tuple))
            if pd.DataFrame(y).isna().any().any():
                raise NAFoundError(name, y)
        # Join dataframes
        true_df, pred_df = pd.DataFrame(y_true), pd.DataFrame(y_pred)
        if how == "all":
            self._true, self._pred = true_df, pred_df
            return
        if on is not None:
            true_df = true_df.set_index(on)
            pred_df = pred_df.set_index(on)
        all_df = true_df.join(pred_df, how="inner", lsuffix=self._A, rsuffix=self._P)
        # Register values
        self._true = all_df.loc[:, [f"{col}{self._A}" for col in true_df.columns]]
        self._pred = all_df.loc[:, [f"{col}{self._P}" for col in pred_df.columns]]

    def score(self, metric=None, metrics="RMSLE"):
        """
        Calculate score with specified metric.

        Args:
            metric (str or None): ME, MAE, MSE, MSLE, MAPE, RMSE, RMSLE, R2 or None (use @metrics)
            metrics (str): alias of @metric

        Raises:
            UnExpectedValueError: un-expected metric was applied
            ValueError: ME was selected as metric when the targets have multiple columns

        Returns:
            float: score with the metric

        Note:
            ME: maximum residual error
            MAE: mean absolute error
            MSE: mean square error
            MSLE: mean squared logarithmic error
            MAPE: mean absolute percentage error
            RMSE: root mean squared error
            RMSLE: root mean squared logarithmic error
            R2: the coefficient of determination

        Note:
            When @metric is None, @metrics will be used as @metric. Default value is "RMSLE".
        """
        metric = (metric or metrics).upper()
        # Check metric name
        if metric not in self._METRICS_DICT:
            raise UnExpectedValueError("metric", metric, candidates=list(self._METRICS_DICT.keys()))
        # Calculate score
        try:
            return float(self._METRICS_DICT[metric][0](self._true.values, self._pred.values))
        except ValueError:
            raise ValueError(
                f"When the targets have multiple columns or negative values, we cannot select {metric}.") from None

    @classmethod
    def metrics(cls):
        """
        Return the list of metric names.

        Returns:
            list[str]: list of metric names
        """
        return list(cls._METRICS_DICT.keys())

    @classmethod
    def smaller_is_better(cls, metric=None, metrics="RMSLE"):
        """
        Whether smaller value of the metric is better or not.

        Args:
            metric (str or None): ME, MAE, MSE, MSLE, MAPE, RMSE, RMSLE, R2 or None (use @metrics)
            metrics (str): alias of @metric

        Returns:
            bool: whether smaller value is better or not
        """
        metric = (metric or metrics).upper()
        # Check metric name
        if metric not in cls._METRICS_DICT:
            raise UnExpectedValueError("metric", metric, candidates=list(cls._METRICS_DICT.keys()))
        return cls._METRICS_DICT[metric][1]

    @classmethod
    def best_one(cls, candidate_dict, **kwargs):
        """
        Select the best one with scores.

        Args:
            candidate_dict (dict[object, float]): scores of candidates
            kwargs: keyword arguments of Evaluator.smaller_is_better()

        Returns:
            tuple(object, float): the best one and its score
        """
        comp_f = {True: min, False: max}[cls.smaller_is_better(**kwargs)]
        return comp_f(candidate_dict.items(), key=lambda x: x[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.error import UnExpectedReturnValueError, NotIncludedError, deprecate
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.feature_engineer import _FeatureEngineer
from covsirphy._deprecated.param_elastic_net import _ParamElasticNetRegressor
from covsirphy._deprecated.param_decision_tree import _ParamDecisionTreeRegressor
from covsirphy._deprecated.param_lightgbm import _ParamLightGBMRegressor
from covsirphy._deprecated.param_svr import _ParamSVRegressor


class RegressionHandler(Term):
    """
    Handle regressors to predict parameter values of ODE models.
    With .fit() method, the best regressor will be selected based on the scores with test dataset.

    Args:
        data (pandas.DataFrame):
            Index
                Date (pandas.Timestamp): observation date
            Columns
                - parameter values
                - the number of cases
                - indicators
        model (covsirphy.ModelBase): ODE model
        kwargs: keyword arguments of sklearn.model_selection.train_test_split()

    Note:
        If @seed is included in kwargs, this will be converted to @random_state.

    Note:
        default values regarding sklearn.model_selection.train_test_split() are
        test_size=0.2, random_state=0, shuffle=False.
    """

    @deprecate(old="RegressionHandler", new="AutoMLHandler", version="2.24.0-xi")
    def __init__(self, data, model, **kwargs):
        # ODE parameter values
        Validator(model, "model").subclass(ModelBase)
        self._parameters = model.PARAMETERS[:]
        # Set datasets (create _FeatureEngineer instance)
        Validator(data, "data").dataframe(columns=self._parameters, time_index=True)
        df = data.drop([self.C, self.CI, self.F, self.R, self.S], axis=1, errors="ignore")
        X = df.drop(self._parameters, axis=1)
        Y = df.loc[:, self._parameters]
        self._engineer = _FeatureEngineer(X, Y)
        # Keyword arguments
        self._kwargs = kwargs.copy()
        # All regressors {name: RegressorBase}
        self._reg_dict = {}
        # The best regressor name and determined delay period
        self._best = None
        # Delay period
        self._delay_candidates = []
        # Backward compatibility, version < 2.21.0-alpha
        if "delay" in kwargs:
            self.feature_engineering(tools=["delay"], delay=kwargs["delay"])

    def _convert_delay_value(self, delay):
        """
        Convert delay value to candidate list of delay periods.

        Args:
            delay (int or tuple(int, int) or None): exact (or value range of) delay period [days]

        Raises:
            ValueError: @delay is None

        Returns:
            list[int]: candidates of delay periods
        """
        # Delay period
        if delay is None:
            raise ValueError(
                "@delay must be integer or tuple(int, int) when @tools is None or includes 'delay'.")
        if isinstance(delay, tuple):
            delay_min, delay_max = delay
            Validator(delay_min, "delay[0]").int(value_range=(1, None))
            Validator(delay_max, "delay[1]").int(value_range=(1, None))
            return list(range(delay_min, delay_max + 1))
        return [Validator(delay, "delay").int(value_range=(1, None))]

    def feature_engineering(self, engineering_tools=None, delay=None):
        """
        Perform feature engineering of X dataset.

        Args:
            engineering_tools (list[str]): list of the feature engineering tools or None (all tools)
            delay (int or tuple(int, int) or None): exact (or value range of) delay period [days]

        Raises:
            ValueError: @delay is None when @tools is None or 'delay' is included in @engineering_tools
            NotIncludedError: @delay was not included in @engineering_tools

        Note:
            All tools and names are
            - "elapsed": accurate elapsed days from the last change point of indicators
            - "log": add log-transformed indicator values
            - "delay": add delayed (lagged) variables with @delay (must not be None)

        Note:
            "delay" must be included in the tools because delay is required to create target X.

        Note:
            "delay" will be applied to all indicators, including features created by the other tools.
        """
        # Delay period
        if engineering_tools is None or "delay" in engineering_tools:
            self._delay_candidates = self._convert_delay_value(delay)
        else:
            raise NotIncludedError("delay", "engineering_tools")
        # Tools of feature engineering
        tool_dict = {
            "elapsed": (self._engineer.add_elapsed, {}),
            "log": (self._engineer.log_transform, {}),
            "delay": (self._engineer.apply_delay, {"delay_values": self._delay_candidates}),
        }
        all_tools = list(tool_dict.keys())
        selected_tools = Validator(engineering_tools, "tools").sequence(default=all_tools, candidates=all_tools)
        # Perform feature engineering
        for name in selected_tools:
            method, arg_dict = tool_dict[name]
            method(**arg_dict)

    def fit(self, metric, regressors=None):
        """
        Fit regressors and select the best regressor based on the scores with test dataset.

        Args:
            metric (str): metric name to select the best regressor
            regressor (list[str]): list of regressors selected from en, dt, lgbm, svr (refer to note)

        Raises:
            ValueError: un-expected parameter values were predicted by all regressors, out of range (0, 1)

        Returns:
            float: the best score

        Note:
            All regressors are here.
            - "en": Indicators -> Parameters with Elastic Net
            - "dt": Indicators -> Parameters with Decision Tree Regressor
            - "lgbm": Indicators -> Parameters with Light Gradient Boosting Machine Regressor
            - "svr": Indicators -> Parameters with Epsilon-Support Vector Regressor
        """
        data_dict = self._engineer.split(**self._kwargs)
        # Select regressors
        all_reg_dict = {
            "en": _ParamElasticNetRegressor,
            "dt": _ParamDecisionTreeRegressor,
            "lgbm": _ParamLightGBMRegressor,
            "svr": _ParamSVRegressor,
        }
        sel_regressors = [
            reg for (name, reg) in all_reg_dict.items() if regressors is None or name in regressors]
        approach_dict = {reg.DESC: reg(**data_dict) for reg in sel_regressors}
        # Predicted all parameter values must be >= 0
        self._reg_dict = {
            k: v for (k, v) in approach_dict.items()
            if v.predict().ge(0).all().all() and v.predict().le(1).all().all()}
        if not self._reg_dict:
            raise UnExpectedReturnValueError(
                name="ODE parameter values", value=None, plural=True,
                details="Values are out of range (0, 1) with all regressors")
        # Select the best regressor with the metric
        score_dict = {k: v.score_test(metric=metric) for (k, v) in self._reg_dict.items()}
        self._best, score = Evaluator.best_one(score_dict, metric=metric)
        return score

    def to_dict(self, metric):
        """
        Return information regarding the best regressor.

        Args:
            metric (str): metric name to select the best regressor

        Returns:
            dict(str, object): regressor information of the best model, including
                - best (str): description of the selected approach
                - scaler (object): scaler class
                - regressor (object): regressor class
                - alpha (float): alpha value used in Elastic Net regression
                - l1_ratio (float): l1_ratio value used in Elastic Net regression
                - score_name (str): scoring method (specified with @metric or @metrics)
                - score_train (float): score with train dataset
                - score_test (float): score with test dataset
                - dataset (dict[numpy.ndarray]): X_train, X_test, y_train, y_test, X_target
                - intercept (pandas.DataFrame): intercept and coefficients (Index ODE parameters, Columns indicators)
                - coef (pandas.DataFrame): intercept and coefficients (Index ODE parameters, Columns indicators)
                - delay (list[int]): list of delay period [days]
        """
        fit_dict = {"best": self._best}
        fit_dict.update(self._reg_dict[self._best].to_dict(metric=metric))
        fit_dict["delay"] = self._delay_candidates
        return fit_dict

    def predict(self):
        """
        Predict parameter values of the ODE model using the best regressor.

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): future dates
                Columns
                    (float): parameter values (4 digits)
        """
        return self._reg_dict[self._best].predict()

    def pred_actual_plot(self, metric, filename=None):
        """
        Create a scatter plot (predicted vs. actual parameter values).

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()
            filename (str): filename of the figure or None (display)
        """
        return self._reg_dict[self._best].pred_actual_plot(metric=metric, filename=filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from io import BytesIO, StringIO
import urllib
from zipfile import ZipFile
import pandas as pd
import requests
from covsirphy.util.term import Term
from covsirphy._deprecated.db_base import _RemoteDatabase


class _COVID19dh(_RemoteDatabase):
    """
    Access COVID-19 Data Hub.

    Args:
        filename (str): CSV filename to save records
        iso3 (str or None): ISO3 code of the country which must be included in the dataset or None (all available countries)
    """
    # Citation
    CITATION = '(Secondary source)' \
        ' Guidotti, E., Ardia, D., (2020), "COVID-19 Data Hub",' \
        ' Journal of Open Source Software 5(51):2376, doi: 10.21105/joss.02376.'
    # Column names and data types
    # {"name in database": "name defined in Term class"}
    _OXCGRT_COLS_RAW = [
        "school_closing",
        "workplace_closing",
        "cancel_events",
        "gatherings_restrictions",
        "transport_closing",
        "stay_home_restrictions",
        "internal_movement_restrictions",
        "international_movement_restrictions",
        "information_campaigns",
        "testing_policy",
        "contact_tracing",
        "stringency_index",
    ]
    OXCGRT_VARS = [v.capitalize() for v in _OXCGRT_COLS_RAW]
    COL_DICT = {
        "date": Term.DATE,
        "administrative_area_level_1": Term.COUNTRY,
        "administrative_area_level_2": Term.PROVINCE,
        "tests": Term.TESTS,
        "confirmed": Term.C,
        "deaths": Term.F,
        "recovered": Term.R,
        "population": Term.N,
        "iso_alpha_3": Term.ISO3,
        **{v: v.capitalize() for v in _OXCGRT_COLS_RAW},
    }

    def download(self, verbose):
        """
        Download the dataset from the server and set the list of primary sources.

        Args:
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by the first values of self.COL_DICT.values()

        Note:
            If @verbose is 2, detailed citation list will be shown when downloading.
            If @verbose is 1, how to show the list will be explained.
        """

        # Download datasets
        if verbose:
            print("Retrieving datasets from COVID-19 Data Hub https://covid19datahub.io/")
        df = self._download()
        # Perform pre-processing
        df = self._preprocessing(df)
        # Show citation list
        if verbose:
            if isinstance(verbose, int) and verbose >= 2:
                print("\nDetailed citaition list:")
                print(self.primary_list)
            else:
                print("\tPlease set verbose=2 to see the detailed citation list.")
        return df

    def _download(self):
        """
        Retrieve dataset and citation list from COVID-19 Data Hub.

        Returns:
            pandas.DataFrame: raw dataset
        """
        if self._iso3 is None:
            url = "https://storage.covid19datahub.io/level/1.csv.zip"
        else:
            url = f"https://storage.covid19datahub.io/country/{self._iso3}.csv.zip"
        raw = self._read_csv(url, col_dict=None, date="date", date_format="%Y-%m-%d")
        raw = raw.groupby("id").ffill()
        # Remove city-level data
        return raw.loc[raw["administrative_area_level_3"].isna()]

    def primary(self):
        """
        Retrieve citation list from COVID-19 Data Hub.

        Returns:
            str: the list of primary sources
        """
        c_url = "https://storage.covid19datahub.io/src.csv"
        try:
            cite = pd.read_csv(c_url, storage_options={"User-Agent": "Mozilla/5.0"})
        except TypeError:
            # When pandas < 1.2: note that pandas v1.2.0 requires Python >= 3.7.1
            r = requests.get(url=c_url, headers={"User-Agent": "Mozilla/5.0"})
            cite = pd.read_csv(StringIO(r.text))
        if self._iso3 is not None:
            cite = cite.loc[cite["iso_alpha_3"] == self._iso3]
        cite = cite.loc[cite["data_type"].isin(self.COL_DICT.keys())]
        cite = cite.loc[:, ["title", "year", "url"]].drop_duplicates(subset="title")
        cite = cite.sort_values(["year", "url"], ascending=[False, True])
        series = cite.apply(lambda x: f"{x[0]} ({x[1]}), {x[2]}", axis=1)
        return "\n".join(series.tolist())

    def _preprocessing(self, raw):
        """
        Perform pre-processing with the raw dataset.

        Args:
            raw (pandas.DataFrame):
                Index
                    reset index
                Columns
                    Refer to COVID-19 Data Hub homepage.
                    https://covid19datahub.io/articles/doc/data.html

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date: observation date
                    - Country: country/region name
                    - Province: province/prefecture/state name
                    - Confirmed: the number of confirmed cases
                    - Infected: the number of currently infected cases
                    - Fatal: the number of fatal cases
                    - Recovered: the number of recovered cases
                    - School_closing
                    - Workplace_closing
                    - Cancel_events
                    - Gatherings_restrictions
                    - Transport_closing
                    - Stay_home_restrictions
                    - Internal_movement_restrictions
                    - International_movement_restrictions
                    - Information_campaigns
                    - Testing_policy
                    - Contact_tracing
                    - Stringency_index

        Note:
            Data types are not confirmed.
        """
        df = raw.copy()
        c, p, iso = "administrative_area_level_1", "administrative_area_level_2", "iso_alpha_3"
        # Country
        df[c] = df[c].replace(
            {
                # COD
                "Congo, the Democratic Republic of the": "Democratic Republic of the Congo",
                # COG
                "Congo": "Republic of the Congo",
                # KOR
                "Korea, South": "South Korea",
            }
        )
        # Set 'Others' as the country name of cruise ships
        ships = ["Diamond Princess", "Costa Atlantica", "Grand Princess", "MS Zaandam"]
        for ship in ships:
            df.loc[df[c] == ship, [c, p, iso]] = [self.OTHERS, ship, self.OTHERS]
        return df

    def _read_csv(self, filepath_or_buffer, col_dict, date="date", date_format="%Y-%m-%d"):
        """Read CSV data.

        Args:
            filepath_or_buffer (str, path object or file-like object): file path or URL
            col_dict (dict[str, str] or None): dictionary to convert column names or None (not perform conversion)
            date (str): column name of date
            date_format (str): format of date column, like %Y-%m-%d
        """
        read_dict = {
            "header": 0, "usecols": None if col_dict is None else list(col_dict.keys()),
            "parse_dates": None if date is None else [date], "date_parser": lambda x: pd.datetime.strptime(x, date_format)
        }
        try:
            df = pd.read_csv(filepath_or_buffer, **read_dict)
        except urllib.error.HTTPError:
            try:
                df = pd.read_csv(filepath_or_buffer, storage_options={"User-Agent": "Mozilla/5.0"}, **read_dict)
            except TypeError:
                # When pandas < 1.2: note that pandas v1.2.0 requires Python >= 3.7.1
                r = requests.get(url=filepath_or_buffer, headers={"User-Agent": "Mozilla/5.0"})
                z = ZipFile(BytesIO(r.content))
                with z.open(z.namelist()[0], "r") as fh:
                    df = pd.read_csv(BytesIO(fh.read()), **read_dict)
        return df if col_dict is None else df.rename(columns=col_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
import warnings
import numpy as np
from datetime import timedelta
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _ComplementHandler(Term):
    """
    Complement records, if necessary.

    Args:
        recovery_period (int): expected value of recovery period [days]
        interval (int): expected update interval of the number of recovered cases [days]
        max_ignored (int): Max number of recovered cases to be ignored [cases]
        max_ending_unupdated (int): Max number of days to apply full complement, where max recovered cases are not updated [days]
        upper_limit_days (int): maximum number of valid partial recovery periods [days]
        lower_limit_days (int): minimum number of valid partial recovery periods [days]
        upper_percentage (float): fraction of partial recovery periods with value greater than upper_limit_days
        lower_percentage (float): fraction of partial recovery periods with value less than lower_limit_days

    Note:
        To add new complement solutions, we need to update cls.STATUS_NAME_DICT and self._protocol().
        Status names with high scores will be prioritized when status code will be determined.
        Status code: 'fully complemented recovered data' and so on as noted in self.run() docstring.
    """
    # Column names
    RAW_COLS = [Term.C, Term.F, Term.R]
    MONOTONIC_CONFIRMED = "Monotonic_confirmed"
    MONOTONIC_FATAL = "Monotonic_fatal"
    MONOTONIC_RECOVERED = "Monotonic_recovered"
    FULL_RECOVERED = "Full_recovered"
    PARTIAL_RECOVERED = "Partial_recovered"
    RECOVERED_COLS = [MONOTONIC_RECOVERED, FULL_RECOVERED, PARTIAL_RECOVERED]
    SHOW_COMPLEMENT_FULL_COLS = [
        MONOTONIC_CONFIRMED, MONOTONIC_FATAL, *RECOVERED_COLS]
    # Kind of complement: {score: name}
    STATUS_NAME_DICT = {
        1: "sorting",
        2: "monotonic increasing",
        3: "partially",
        4: "fully",
    }

    def __init__(self, recovery_period, interval, max_ignored, max_ending_unupdated,
                 upper_limit_days, lower_limit_days, upper_percentage, lower_percentage):
        # Arguments for complement
        self.recovery_period = Validator(recovery_period, "recovery_period").int(value_range=(1, None))
        self.interval = Validator(interval, "interval").int(value_range=(0, None))
        self.max_ignored = Validator(max_ignored, "max_ignored").int(value_range=(1, None))
        self.max_ending_unupdated = Validator(max_ending_unupdated, "max_ending_unupdated").int(value_range=(0, None))
        self.upper_limit_days = Validator(upper_limit_days, "upper_limit_days").int(value_range=(0, None))
        self.lower_limit_days = Validator(lower_limit_days, "lower_limit_days").int(value_range=(0, None))
        self.upper_percentage = Validator(upper_percentage, "upper_percentage").float(value_range=(0, 100))
        self.lower_percentage = Validator(lower_percentage, "lower_percentage").float(value_range=(0, 100))
        self.complement_dict = None

    def _protocol(self):
        """
        Return the list of complement solutions and scores.

        Returns:
            list[str/None, function, int]: nested list of variables to be updated, methods and scores
        """
        return [
            (None, self._pre_processing, 0),
            (self.C, functools.partial(self._monotonic, variable=self.C), 2),
            (self.F, functools.partial(self._monotonic, variable=self.F), 2),
            (self.R, functools.partial(self._monotonic, variable=self.R), 2),
            (self.R, self._recovered_full, 4),
            (self.R, self._recovered_sort, 1),
            (self.R, functools.partial(self._monotonic, variable=self.R), 2),
            (self.R, self._recovered_partial, 3),
            (self.R, self._recovered_partial_ending, 3),
            (self.R, self._recovered_sort, 1),
            (self.R, functools.partial(self._monotonic, variable=self.R), 2),
            (self.R, self._recovered_sort, 1),
            (self.C, self._confirmed_over_fr, 1),
            (None, self._post_processing, 0)
        ]

    def run(self, data):
        """
        Perform complement.

        Args:
            data (pandas.DataFrame): Subset of records
                Index
                    reset index
                Columns
                    Date (pd.Timestamp): Observation date
                    Confirmed (int): the number of confirmed cases
                    Fatal (int): the number of fatal cases
                    Recovered (int): the number of recovered cases
                    The other columns will be ignored

        Returns:
            tuple(pandas.DataFrame, str, dict):
                pandas.DataFrame
                    Index
                        reset index
                    Columns
                        Date (pd.Timestamp): Observation date
                        Confirmed (int): the number of confirmed cases
                        Fatal (int): the number of fatal cases
                        Recovered (int): the number of recovered cases
                str: status code: will be selected from
                    - '' (not complemented)
                    - 'monotonic increasing complemented confirmed data'
                    - 'monotonic increasing complemented fatal data'
                    - 'monotonic increasing complemented recovered data'
                    - 'fully complemented recovered data'
                    - 'partially complemented recovered data'
                dict[str, bool]: status for each complement type, keys are
                    - Monotonic_confirmed
                    - Monotonic_fatal
                    - Monotonic_recovered
                    - Full_recovered
                    - Partial_recovered
        """
        Validator(data, "data").dataframe(columns=[self.DATE, *self.RAW_COLS])
        # Initialize
        after_df = data.copy()
        status_dict = dict.fromkeys(self.RAW_COLS, 0)
        self.complement_dict = dict.fromkeys(self.SHOW_COMPLEMENT_FULL_COLS, False)
        # Perform complement one by one
        for (variable, func, score) in self._protocol():
            before_df, after_df = after_df.copy(), func(after_df)
            if after_df.equals(before_df) or variable is None:
                continue
            status_dict[variable] = max(status_dict[variable], score)
        # Create status code
        status_list = [
            f"{self.STATUS_NAME_DICT[score]} complemented {v.lower()} data" for (v, score) in status_dict.items() if score]
        return (after_df.convert_dtypes(), " and \n".join(status_list), self.complement_dict)

    def _pre_processing(self, data):
        """
        Select Confirmed/Fatal/Recovered class from the dataset.

        Args:
            data (pandas.DataFrame): Subset of records
                Index
                    reset index
                Columns
                    Date, Confirmed, Fatal, Recovered (the others will be ignored)

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered
        """
        sel_invalid_R = data[self.C] - data[self.F] < data[self.R]
        data.loc[sel_invalid_R, self.R] = data[self.C] - data[self.F]
        data.loc[sel_invalid_R, self.CI] = data[self.C] - data[self.F] - data[self.R]
        return data.set_index(self.DATE).loc[:, self.RAW_COLS]

    def _post_processing(self, data):
        """
        Select Confirmed/Fatal/Recovered class from the dataset.

        Args:
            data (pandas.DataFrame):
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index
                    reset index
                Columns
                    Date (pandas.TimeStamp), Confirmed, Fatal, Recovered (int)
        """
        return data.loc[:, self.RAW_COLS].astype(np.int64).reset_index()

    def _monotonic(self, data, variable):
        """
        Force the variable show monotonic increasing.

        Args:
            df (pandas.DataFrame)
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered
            variable (str): variable name to show monotonic increasing

        Returns:
            pandas.DataFrame: complemented records
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered
        """
        df = data.astype(np.float64)
        warnings.simplefilter("ignore", UserWarning)
        # Whether complement is necessary or not
        if df[variable].is_monotonic_increasing:
            return df
        # Complement
        decreased_dates = df[df[variable].diff() < 0].index.tolist()
        for date in decreased_dates:
            # Raw value on the decreased date
            raw_last = df.loc[date, variable]
            # Extrapolated value on the date
            series = df.loc[:date, variable]
            series.iloc[-1] = None
            try:
                series = series.interpolate().ffill()
            except ValueError:
                series.ffill(inplace=True)
            # Reduce values to the previous date
            df.loc[:date, variable] = series * raw_last / series.iloc[-1]
            df[variable] = df[variable].fillna(0)
        df[variable] = df[variable].apply(np.ceil).astype(np.int64)
        self.complement_dict[f"Monotonic_{variable.lower()}"] = True
        return df

    def _validate_recovery_period(self, country_df):
        """
        Calculates and validates recovery period for specific country
        as an additional condition in order to apply full complement or not

        Args:
            country_df (pandas.DataFrame)
                Index
                    reset_index
                Columns
                    Date, Confirmed, Recovered, Fatal

        Returns:
            bool: true if recovery period is within valid range or false otherwise

        Note: Passed argument df corresponds to specific country's df
              upper_limit_days has default value of 3 months (90 days)
              lower_limit_days has default value of 1 week (7 days)
        """
        df = country_df.copy()
        # Calculate "Confirmed - Fatal"
        df["diff"] = df[self.C] - df[self.F]
        df = df.loc[:, ["diff", self.R]]
        # Calculate how many days passed to reach the number of cases
        df = df.unstack().reset_index()
        df.columns = ["Variable", "Date", "Number"]
        df["Days"] = (df[self.DATE] - df[self.DATE].min()).dt.days
        # Calculate partial recovery periods
        df = df.pivot_table(values="Days", index="Number", columns="Variable")
        df = df.interpolate(limit_area="inside").dropna().astype(np.int64)
        df["Elapsed"] = df[self.R] - df["diff"]
        df = df.loc[df["Elapsed"] > 0]
        # Check partial recovery periods
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        per_up = (df["Elapsed"] > self.upper_limit_days).sum() / len(df)
        per_lw = (df["Elapsed"] < self.lower_limit_days).sum() / len(df)
        return per_up < self.upper_percentage and per_lw < self.lower_percentage

    def _recovered_full(self, df):
        """
        Estimate the number of recovered cases with the value of recovery period.

        Args:
            df (pandas.DataFrame)
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered
        """
        # Whether complement is necessary or not
        if df[self.R].max() > self.max_ignored:
            # Necessary if sum of recovered is more than 99%
            # or less than 1% of sum of recovered minus infected when out-breaking
            sel_C1 = df[self.C] > self.max_ignored
            sel_R1 = df[self.R] > self.max_ignored
            sel_2 = df[self.C].diff().diff().rolling(14).mean() > 0
            cf_diff = (df[self.C] - df[self.F]).rolling(14).mean()
            sel_3 = df[self.R] < 0.01 * cf_diff
            s_df = df.loc[sel_C1 & sel_R1 & sel_2 & sel_3]
            if s_df.empty and self._validate_recovery_period(df):
                return df
        # Estimate recovered records
        df[self.R] = (df[self.C] - df[self.F]).shift(periods=self.recovery_period, freq="D")
        self.complement_dict[self.FULL_RECOVERED] = True
        return df

    def _recovered_partial_ending(self, df):
        """
        If ending recovered values do not change for more than applied 'self.max_ending_unupdated' days
        after reached 'self.max_ignored' cases, apply either previous diff() (with some variance)
        or full complement only to these ending unupdated values and keep the previous valid ones.
        _recovered_partial() does not handle well such values, because
        they are not in-between values and interpolation generates only similar values,
        but small compared to confirmed cases.

        Args:
            df (pandas.DataFrame)
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Note: _recovered_partial_ending() must always be called
              after _recovered_partial()
        """
        # Whether complement is necessary or not
        r_max = df[self.R].max()
        sel_0 = (df[self.R] == r_max).sum() > self.max_ending_unupdated
        if not (r_max and sel_0):
            # full complement will be handled in _recovered_full()
            return df
        # Complement any ending unupdated values that are not updated
        # for more than max_ending_unupdated days,
        # by keeping and propagating forward previous valid diff()
        # min_index: index for first ending max R reoccurrence
        min_index = df[self.R].idxmax() + timedelta(days=1)
        first_value = df.loc[min_index, self.R]
        df_ending = df.copy()
        df_ending.loc[df_ending.duplicated([self.R], keep="first"), self.R] = None
        diff_series = df_ending[self.R].diff().ffill().fillna(0).astype(np.int64)
        diff_series.loc[diff_series.duplicated(keep="last")] = None
        diff_series.interpolate(method="linear", inplace=True, limit_direction="both")
        df.loc[min_index:, self.R] = first_value + diff_series[min_index:].cumsum()
        # Check if the ending complement is valid (too large recovered ending values)
        # If the validity check fails, then fully complement these ending values
        sel_C1 = df[self.C] > self.max_ignored
        sel_R1 = df[self.R] > self.max_ignored
        # check all values one-by-one, no rolling window
        cf_diff = df[self.C] - df[self.F]
        sel_limit = df[self.R] > 0.99 * cf_diff
        s_df_1 = df.loc[sel_C1 & sel_R1 & sel_limit]
        if not s_df_1.empty:
            df.loc[min_index:, self.R] = (
                df[self.C] - df[self.F]).shift(periods=self.recovery_period, freq="D").loc[min_index:]
        self.complement_dict[self.PARTIAL_RECOVERED] = True
        return df

    def _recovered_partial(self, df):
        """
        If recovered values do not change for more than applied 'self.interval' days
        after reached 'self.max_ignored' cases, interpolate the recovered values.

        Args:
            df (pandas.DataFrame)
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered
        """
        # Whether complement is necessary or not
        series = df.loc[df[self.R] > self.max_ignored, self.R]
        max_frequency = series.value_counts().max()
        if max_frequency <= self.interval:
            return df
        # Complement in-between recovered values when confirmed > max_ignored
        sel_C = df[self.C] > self.max_ignored
        sel_duplicate = df.duplicated([self.R], keep="first")
        df.loc[sel_C & sel_duplicate, self.R] = None
        df[self.R].interpolate(method="linear", inplace=True, limit_direction="both")
        df[self.R] = df[self.R].fillna(method="bfill")
        self.complement_dict[self.PARTIAL_RECOVERED] = True
        return df

    def _recovered_sort(self, df):
        """
        Sort the absolute values of recovered data.

        Args:
            df (pandas.DataFrame)
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Note:
            _recovered_sort() must always be called after _recovered_partial_ending()
        """
        df.loc[:, self.R] = sorted(df[self.R].abs())
        df[self.R] = df[self.R].interpolate(method="time").fillna(0).round()
        return df

    def _confirmed_over_fr(self, df):
        """
        Force Confirmed > Fatal + Recovered.

        Args:
            df (pandas.DataFrame)
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered

        Returns:
            pandas.DataFrame: complemented records
                Index
                    Date (pandas.TimeStamp)
                Columns
                    Confirmed, Fatal, Recovered
        """
        df[self.C] = df[[self.C, self.F, self.R]].apply(lambda x: max(x[0], x[1] + x[2]), axis=1)
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import defaultdict
from datetime import datetime, timedelta
import logging
import math
import warnings
import country_converter as coco
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator


class Term(object):
    """
    Term definition.
    """
    # Variables of SIR-derived model
    N = "Population"
    S = "Susceptible"
    C = "Confirmed"
    CI = "Infected"
    F = "Fatal"
    R = "Recovered"
    FR = "Fatal or Recovered"
    E = "Exposed"
    W = "Waiting"
    # PCR tests
    TESTS = "Tests"
    TESTS_DIFF = "Tests_diff"
    # Severity
    MODERATE = "Moderate"
    SEVERE = "Severe"
    # Vaccination
    VAC = "Vaccinations"
    V = "Vaccinated"
    V_ONCE = f"{V}_once"
    V_FULL = f"{V}_full"
    VAC_BOOSTERS = f"{VAC}_boosters"
    PRODUCT = "Product"
    # Column names
    DATE = "Date"
    START = "Start"
    END = "End"
    T = "Elapsed"
    TS = "t"
    TAU = "tau"
    COUNTRY = "Country"
    ISO3 = "ISO3"
    PROVINCE = "Province"
    CITY = "City"
    STEP_N = "step_n"
    Y0_DICT = "y0_dict"
    PARAM_DICT = "param_dict"
    ID = "ID"
    _PH = "Phase_ID"
    _SIRF = [S, CI, R, F]
    AREA_COLUMNS = [COUNTRY, PROVINCE]
    STR_COLUMNS = [DATE, *AREA_COLUMNS]
    COLUMNS = [*STR_COLUMNS, C, CI, F, R]
    NLOC_COLUMNS = [DATE, C, CI, F, R]
    SUB_COLUMNS = [DATE, C, CI, F, R, S]
    VALUE_COLUMNS = [C, CI, F, R]
    FIG_COLUMNS = [CI, F, R, FR, V, E, W]
    MONO_COLUMNS = [C, F, R]
    AREA_ABBR_COLS = [ISO3, *AREA_COLUMNS]
    DSIFR_COLUMNS = [DATE, S, CI, F, R]
    # Date format: 22Jan2020 etc.
    DATE_FORMAT = "%d%b%Y"
    DATE_FORMAT_DESC = "DDMmmYYYY"
    # Separator of country and province
    SEP = "/"
    # EDA
    RATE_COLUMNS = [
        "Fatal per Confirmed",
        "Recovered per Confirmed",
        "Fatal per (Fatal or Recovered)"
    ]
    # Optimization
    A = "_actual"
    P = "_predicted"
    ACTUAL = "Actual"
    FITTED = "Fitted"
    # Phase name
    SUFFIX_DICT = defaultdict(lambda: "th")
    SUFFIX_DICT.update({1: "st", 2: "nd", 3: "rd"})
    # Summary of phases
    TENSE = "Type"
    PAST = "Past"
    FUTURE = "Future"
    INITIAL = "Initial"
    ODE = "ODE"
    RT = "Rt"
    RT_FULL = "Reproduction number"
    TRIALS = "Trials"
    RUNTIME = "Runtime"
    # Scenario analysis
    PHASE = "Phase"
    SERIES = "Scenario"
    MAIN = "Main"
    # Flag
    NA = "-"
    OTHERS = "Others"

    @classmethod
    def num2str(cls, num):
        """
        Convert numbers to 1st, 2nd etc.

        Args:
            num (int): number

        Returns:
            str
        """
        num = Validator(num, "num").int(value_range=(0, None))
        q, mod = divmod(num, 10)
        suffix = "th" if q % 10 == 1 else cls.SUFFIX_DICT[mod]
        return f"{num}{suffix}"

    @staticmethod
    def str2num(string, name="phase names"):
        """
        Convert 1st to 1 and so on.

        Args:
            string (str): like 1st, 2nd, 3rd,...
            name (str): name of the string

        Returns:
            int
        """
        try:
            return int(string[:-2])
        except ValueError as e:
            raise ValueError(
                f"Examples of {name} are 0th, 1st, 2nd..., but {string} was applied."
            ) from e

    @staticmethod
    @deprecate(".negative_exp()", version="2.25.0-mu")
    def negative_exp(x, a, b):
        """
        Negative exponential function f(x) = A exp(-Bx).

        Args:
            x (float): x values
            a (float): the first parameters of the function
            b (float): the second parameters of the function

        Returns:
            float
        """
        return a * np.exp(-b * x)

    @staticmethod
    @deprecate(".linear()", version="2.25.0-mu")
    def linear(x, a, b):
        """
        Linear function f(x) = A x + b.

        Args:
            x (float): x values
            a (float): the first parameter of the function
            b (float): the second parameter of the function

        Returns:
            float
        """
        return a * x + b

    @staticmethod
    @deprecate(".flatten()", new="Validator.sequence(flatten=True)", version="2.25.0-mu")
    def flatten(nested_list, unique=True):
        """
        Flatten the nested list.

        Args:
            nested_list (list[list[object]]): nested list
            unique (bool): if True, only unique values will remain

        Returns:
            list[object]
        """
        flattened = sum(nested_list, [])
        return list(set(flattened)) if unique else flattened

    @classmethod
    @deprecate(".divisors()", version="2.25.0-mu")
    def divisors(cls, value):
        """
        Return the list of divisors of the value.

        Args:
            value (int): target value

        Returns:
            list[int]: the list of divisors
        """
        value = Validator(value).int(value_range=(1, None))
        return [i for i in range(1, value + 1) if value % i == 0]

    @classmethod
    @deprecate(".date_obj()", version="2.19.1-gamma-fu5")
    def date_obj(cls, date_str=None, default=None):
        """
        Convert a string to a datetime object.

        Args:
            date_str (str or None, optional): string, like 22Jan2020
            default (datetime.datetime or None, optional): default value to return

        Returns:
            datetime.datetime or None

        Note:
            If @date_str is None, returns @default value
        """
        if date_str is None:
            return default
        return datetime.strptime(date_str, cls.DATE_FORMAT)

    @classmethod
    @deprecate(".date_change()", version="2.25.0-nu")
    def _ensure_date(cls, target, name="date", default=None):
        """
        Ensure the format of the string.

        Args:
            target (str or pandas.Timestamp): string to ensure
            name (str): argument name of the string
            default (pandas.Timestamp or None): default value to return

        Returns:
            pandas.Timestamp or None: as-is the target or default value
        """
        if target is None:
            return default
        if isinstance(target, pd.Timestamp):
            return target.replace(hour=0, minute=0, second=0, microsecond=0)
        try:
            return pd.to_datetime(target).replace(hour=0, minute=0, second=0, microsecond=0)
        except ValueError as e:
            raise ValueError(f"{name} was not recognized as a date, {target} was applied.") from e

    @classmethod
    @deprecate(".date_change()", version="2.25.0-mu")
    def date_change(cls, date_str, days=0):
        """
        Return @days days ago or @days days later.

        Args:
            date_str (str): today
            days (int): (negative) days ago or (positive) days later

        Returns:
            str: the date
        """
        if not isinstance(days, int):
            raise TypeError(
                f"@days must be integer, but {type(days)} was applied.")
        date = cls._ensure_date(date_str) + timedelta(days=days)
        return date.strftime(cls.DATE_FORMAT)

    @classmethod
    @deprecate(".tomorrow()", version="2.25.0-mu")
    def tomorrow(cls, date_str):
        """
        Tomorrow of the date.

        Args:
            date_str (str): today

        Returns:
            str: tomorrow
        """
        return cls.date_change(date_str, days=1)

    @classmethod
    @deprecate(".yesterday()", version="2.25.0-mu")
    def yesterday(cls, date_str):
        """
        Yesterday of the date.

        Args:
            date_str (str): today

        Returns:
            str: yesterday
        """
        return cls.date_change(date_str, days=-1)

    @classmethod
    @deprecate(".steps()", version="2.25.0-mu")
    def steps(cls, start_date, end_date, tau):
        """
        Return the number of days (round up).

        Args:
            start_date (str): start date, like 01Jan2020
            end_date (str): end date, like 01Jan2020
            tau (int): tau value [min]
        """
        sta = cls._ensure_date(start_date)
        end = cls._ensure_date(end_date)
        tau = Validator(tau, "tau").tau(default=None)
        return math.ceil((end - sta) / timedelta(minutes=tau))

    @classmethod
    def _to_iso3(cls, name):
        """Convert country name(s) to ISO3 codes.

        Args:
            name (str or list[str] or None): country name(s)

        Returns:
            list[str]: ISO3 code(s)

        Note:
            "UK" will be converted to "GBR".

        Note:
            When the country was not found or None, it will not be converted.

        Examples:
            >>> Term._to_iso3("Japan")
            ['JPN']
            >>> Term._to_iso3("UK")
            ['GBR']
            >>> Term._to_iso3("Moon")
            ['Moon']
            >>> Term._to_iso3(None)
            ['---']
            >>> Term._to_iso3(["Japan", "UK", "Moon", None])
            ['JPN', 'GBR', 'Moon', '---']
        """
        logging.basicConfig(level=logging.CRITICAL)
        warnings.simplefilter("ignore", FutureWarning)
        names = [name] if (isinstance(name, str) or name is None) else name
        code_dict = {"UK": "GBR", None: cls.NA * 3, }
        code_dict.update({elem: coco.convert(elem, to="ISO3", not_found=elem) for elem in set(names) - set(code_dict)})
        return [code_dict[elem] for elem in names]

    def _country_information(self):
        """Return the raw data of country_converter library raw data as a dataframe.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns:
                    - name_short: standard or short names
                    - ISO2: ISO2 codes
                    - ISO3: ISO3 codes
                    - Continent: continent names
                    - the other columns listed in country_converter library homepage.

        Note:
            Refer to https://github.com/konstantinstadler/country_converter
        """
        return coco.CountryConverter().data


class Word(Term):
    @deprecate(old="Word()", new="Term()")
    def __init__(self):
        super().__init__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from pathlib import Path
import warnings
import geopandas as gpd
from matplotlib import pyplot as plt
import pandas as pd
from covsirphy.util.argument import find_args
from covsirphy.util.error import deprecate, SubsetNotFoundError, UnExpectedValueError
from covsirphy.util.term import Term
from covsirphy._deprecated.colored_map import ColoredMap


class CleaningBase(Term):
    """
    Basic class for data cleaning.

    Args:
        filename (str or None): CSV filename of the dataset
        data (pandas.DataFrame or None):
            Index
                reset index
            Columns
                - Date: Observation date
                - ISO3: ISO3 code
                - Country: country/region name (optional)
                - Province: province/prefecture/state name
                - Confirmed: the number of confirmed cases
                - Fatal: the number of fatal cases
                - Recovered: the number of recovered cases
                - Population: population values (optional)
        citation (str or None): citation or None (empty)
        variables (list[str] or None): variables to clean (not including date and location identifiers)

    Note:
        Either @filename (high priority) or @data must be specified.

    Note:
        - If @filename is None, geography information will be saved in "input" directory.
        - If @filename is not None, geography information will be saved in the directory which has the file.
        - The directory of geography information could be changed with .directory property.
    """

    def __init__(self, filename=None, data=None, citation=None, variables=None):
        # Columns of self._raw, self._clean_df and self.cleaned()
        self._raw_cols = [self.DATE, self.ISO3, self.COUNTRY, self.PROVINCE] + (variables or [])
        self._subset_cols = [self.DATE] + (variables or [])
        # Raw data
        self._raw = self._parse_raw(filename, data, self._raw_cols)
        # Data cleaning
        self._cleaned_df = pd.DataFrame(columns=self._raw_cols) if self._raw.empty else self._cleaning()
        # Citation
        self._citation = citation or ""
        # Directory that save the file
        if filename is None:
            self._dirpath = Path("input")
        else:
            Path(filename).parent.mkdir(exist_ok=True, parents=True)
            self._dirpath = Path(filename).resolve().parent

    def _parse_raw(self, filename, data, columns):
        """
        Parse raw data with a CSV file or a dataframe.

        Args:
            filename (str or None): CSV filename of the dataset
            data (pandas.DataFrame or None):
                Index
                    reset index
                Columns
                    columns defined by @required_cols
            columns (list[str]): column names to use

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    columns defined by @required_cols

        Note:
            Either @filename (high priority) or @data must be specified.

        Note:
            If some columns are not included in the dataset, values will be None.
        """
        if filename is None:
            if data is None:
                return pd.DataFrame(columns=columns)
            return data.reindex(columns=columns)
        dtype_dict = {
            self.PROVINCE: "object", "Province/State": "object",
            "key": "object", "key_alpha_2": "object"
        }
        return pd.read_csv(filename, dtype=dtype_dict).reindex(columns=columns)

    @staticmethod
    def _ensure_instance(target, class_obj, name="target"):
        """
        Ensure the target is an instance of the class object.

        Args:
            target (instance): target to ensure
            parent (class): class object
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        s = f"@{name} must be an instance of {class_obj}, but {type(target)} was applied."
        if not isinstance(target, class_obj):
            raise TypeError(s)
        return target

    @staticmethod
    def _ensure_dataframe(target, name="df", time_index=False, columns=None, empty_ok=True):
        """
        Ensure the dataframe has the columns.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            name (str): argument name of the dataframe
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            empty_ok (bool): whether give permission to empty dataframe or not

        Returns:
            pandas.DataFrame:
                Index
                    as-is
                Columns:
                    columns specified with @columns or all columns of @target (when @columns is None)
        """
        if not isinstance(target, pd.DataFrame):
            raise TypeError(f"@{name} must be a instance of (pandas.DataFrame).")
        df = target.copy()
        if time_index and (not isinstance(df.index, pd.DatetimeIndex)):
            raise TypeError(f"Index of @{name} must be <pd.DatetimeIndex>.")
        if not empty_ok and target.empty:
            raise ValueError(f"@{name} must not be a empty dataframe.")
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            cols_str = ", ".join(col for col in columns if col not in df.columns)
            included = ", ".join(df.columns.tolist())
            s1 = f"Expected columns were not included in {name} with {included}."
            raise KeyError(f"{s1} {cols_str} must be included.")
        return df.loc[:, columns]

    @staticmethod
    def _ensure_natural_int(target, name="number", include_zero=False, none_ok=False):
        """
        Ensure a natural (non-negative) number.

        Args:
            target (int or float or str or None): value to ensure
            name (str): argument name of the value
            include_zero (bool): include 0 or not
            none_ok (bool): None value can be applied or not.

        Returns:
            int: as-is the target

        Note:
            When @target is None and @none_ok is True, None will be returned.
            If the value is a natural number and the type was float or string,
            it will be converted to an integer.
        """
        if target is None and none_ok:
            return None
        s = f"@{name} must be a natural number, but {target} was applied"
        try:
            number = int(target)
        except TypeError as e:
            raise TypeError(f"{s} and not converted to integer.") from e
        if number != target:
            raise ValueError(f"{s}. |{target} - {number}| > 0")
        min_value = 0 if include_zero else 1
        if number < min_value:
            raise ValueError(f"{s}. This value is under {min_value}")
        return number

    @classmethod
    def _ensure_date(cls, target, name="date", default=None):
        """
        Ensure the format of the string.

        Args:
            target (str or pandas.Timestamp): string to ensure
            name (str): argument name of the string
            default (pandas.Timestamp or None): default value to return

        Returns:
            pandas.Timestamp or None: as-is the target or default value
        """
        if target is None:
            return default
        if isinstance(target, pd.Timestamp):
            return target.replace(hour=0, minute=0, second=0, microsecond=0)
        try:
            return pd.to_datetime(target).replace(hour=0, minute=0, second=0, microsecond=0)
        except ValueError as e:
            raise ValueError(f"{name} was not recognized as a date, {target} was applied.") from e

    @property
    def raw(self):
        """
        pandas.DataFrame: raw data
        """
        return self._raw

    @raw.setter
    def raw(self, dataframe):
        """
        pandas.DataFrame: raw dataset
        """
        self._raw = self._ensure_dataframe(dataframe, name="dataframe")

    @property
    def directory(self):
        """
        str: directory name to save geography information
        """
        return str(self._dirpath)

    @directory.setter
    def directory(self, name):
        self._dirpath = Path(name)

    @staticmethod
    @deprecate(".load()", version="2.21.0-kappa-fu5")
    def load(urlpath, header=0, columns=None, dtype="object"):
        """
        Load a local/remote file.

        Args:
            urlpath (str or pathlib.Path): filename or URL
            header (int): row number of the header
            columns (list[str]): columns to use
            dtype (str or dict[str]): data type for the dataframe or specified columns

        Returns:
            pd.DataFrame: raw dataset
        """
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        kwargs = {
            "low_memory": False, "dtype": dtype, "header": header, "usecols": columns}
        return pd.read_csv(urlpath, **kwargs)

    def cleaned(self):
        """
        Return the cleaned dataset.

        Note:
            Cleaning method is defined by CleaningBase._cleaning() method.

        Returns:
            pandas.DataFrame: cleaned data
        """
        return self._cleaned_df

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame: cleaned data
        """
        return self._raw.copy()

    @property
    def citation(self):
        """
        str: citation/description of the dataset
        """
        return self._citation

    @citation.setter
    def citation(self, description):
        self._citation = str(description)

    def ensure_country_name(self, country, errors="raise"):
        """
        Ensure that the country name is correct.
        If not, the correct country name will be found.

        Args:
            country (str): country name
            errors (str): 'raise' or 'coerce'

        Returns:
            str: ISO3 code

        Raises:
            SubsetNotFoundError: no records were found for the country and @errors is 'raise'
        """
        iso3 = self._to_iso3(country)[0]
        df = self._cleaned_df.copy()
        if self.ISO3 in df and iso3 in df[self.ISO3].unique():
            return iso3
        self._ensure_dataframe(df, name="the cleaned dataset", columns=[self.COUNTRY])
        selectable_set = set(df[self.COUNTRY].unique())
        # return country name as-is if selectable
        if country in selectable_set:
            return iso3
        if errors == "raise":
            raise SubsetNotFoundError(country=country, country_alias=iso3)

    @deprecate("CleaningBase.iso3_to_country()")
    def iso3_to_country(self, iso3_code):
        """
        Convert ISO3 code to country name if records are available.

        Args:
            iso3_code (str): ISO3 code or country name

        Returns:
            str: country name

        Note:
            If ISO3 codes are not registered, return the string as-si @iso3_code.
        """
        return self.ensure_country_name(iso3_code)

    @deprecate("CleaningBase.country_to_iso3()", new="CleaningBase.ensure_country_name()", version="2.24.0-gamma")
    def country_to_iso3(self, country, check_data=True):
        """
        Convert country name to ISO3 code if records are available.

        Args:
            country (str): country name
            check_data (bool): whether validate the country name with the dataset

        Raises:
            KeyError: ISO3 code of the country is not registered

        Returns:
            str: ISO3 code or "---" (when unknown)
        """
        name = self.ensure_country_name(country) if check_data else country
        iso3 = self._to_iso3(name)[0]
        return self.NA * 3 if iso3 is None or iso3 == country else iso3

    @classmethod
    def area_name(cls, country, province=None):
        """
        Return area name of the province/country.

        Args:
            country (str): country name or ISO3 code
            province (str): province name

        Returns:
            str: area name

        Note:
            If province is None or '-', return country name.
            If not, return the area name, like 'Tokyo/Japan'
        """
        if province in [None, cls.NA]:
            return country
        return f"{province}{cls.SEP}{country}"

    def layer(self, country=None):
        """
        Return the cleaned data at the selected layer.

        Args:
            country (str or None): country name or None (country level data or country-specific dataset)

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                - Country (str): country names
                - Province (str): province names (or removed when country level data)
                - any other columns of the cleaned data

        Raises:
            SubsetNotFoundError: no records were found for the country (when @country is not None)
            KeyError: @country was None, but country names were not registered in the dataset

        Note:
            When @country is None, country level data will be returned.
            When @country is a country name, province level data in the selected country will be returned.
        """
        df = self._cleaned_df.copy()
        for col in [self.ISO3, self.COUNTRY, self.PROVINCE]:
            df[col] = df[col].astype(str, errors="ignore")
        # Country level data
        if country is None:
            if self.PROVINCE in df:
                df = df.loc[df[self.PROVINCE] == self.NA].drop(self.PROVINCE, axis=1)
            return df.reset_index(drop=True)
        # Province level data at the selected country
        self._ensure_dataframe(df, name="the cleaned dataset", columns=[self.PROVINCE])
        iso3 = self.ensure_country_name(country)
        df = df.loc[df[self.ISO3] == iso3]
        if df.empty:
            raise SubsetNotFoundError(country=country, country_alias=iso3) from None
        if self.PROVINCE in df:
            df = df.loc[df[self.PROVINCE] != self.NA]
        return df.reset_index(drop=True)

    def _subset_by_area(self, country, province=None):
        """
        Return subset for the country/province.

        Args:
            country (str): country name
            province (str or None): province name or None (country level data)

        Returns:
            pandas.DataFrame: subset for the country/province, columns are not changed

        Raises:
            SubsetNotFoundError: no records were found for the condition
        """
        # Country level
        if province is None or province == self.NA:
            df = self.layer(country=None)
            iso3 = self.ensure_country_name(country)
            df = df.loc[df[self.ISO3] == iso3]
            return df.reset_index(drop=True)
        # Province level
        df = self.layer(country=country)
        df = df.loc[df[self.PROVINCE] == province]
        if df.empty:
            raise SubsetNotFoundError(country=country)
        return df.reset_index(drop=True)

    def subset(self, country, province=None, start_date=None, end_date=None):
        """
        Return subset with country/province name and start/end date.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    without ISO3, Country, Province column

        Raises:
            SubsetNotFoundError: no records were found for the condition
        """
        iso3 = self.ensure_country_name(country, errors="coerce")
        try:
            df = self._subset_by_area(country=country, province=province)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                country=country, country_alias=iso3, province=province) from None
        df = df.drop([self.COUNTRY, self.ISO3, self.PROVINCE], axis=1, errors="ignore")
        # Subset with Start/end date
        if start_date is None and end_date is None:
            return df.reset_index(drop=True)
        self._ensure_dataframe(df, name="the cleaned dataset", columns=[self.DATE])
        series = df[self.DATE].copy()
        start_obj = self._ensure_date(start_date, default=series.min())
        end_obj = self._ensure_date(end_date, default=series.max())
        df = df.loc[(start_obj <= series) & (series <= end_obj), :]
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=iso3, province=province,
                start_date=start_date, end_date=end_date) from None
        return df.reset_index(drop=True)

    def subset_complement(self, country, **kwargs):
        """
        Return the subset. If necessary, complementing will be performed.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def records(self, country, province=None, start_date=None, end_date=None,
                auto_complement=True, **kwargs):
        """
        Return the subset. If necessary, complementing will be performed.

        Args:
            country (str): country name or ISO3 code
            province (str or None): province name
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            auto_complement (bool): if True and necessary, the number of cases will be complemented
            kwargs: the other arguments of complement

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    without ISO3, Country, Province column
        """
        iso3 = self.ensure_country_name(country)
        subset_arg_dict = {
            "country": country, "province": province, "start_date": start_date, "end_date": end_date}
        if auto_complement:
            with contextlib.suppress(NotImplementedError):
                df, is_complemented = self.subset_complement(
                    **subset_arg_dict, **kwargs)
                if not df.empty:
                    return (df, is_complemented)
        try:
            return (self.subset(**subset_arg_dict), False)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                country=country, country_alias=iso3, province=province,
                start_date=start_date, end_date=end_date) from None

    def countries(self):
        """
        Return names of countries where records are registered.

        Raises:
            KeyError: Country names are not registered in this dataset

        Returns:
            list[str]: list of country names
        """
        df = self._ensure_dataframe(
            self._cleaned_df, name="the cleaned dataset", columns=[self.COUNTRY])
        return list(df[self.COUNTRY].unique())

    def total(self):
        """
        Calculate total values of the cleaned dataset.
        """
        raise NotImplementedError

    def _colored_map(self, title, **kwargs):
        """
        Create global colored map to show the values.

        Args:
            title (str): title of the figure
            kwargs: arguments of ColoredMap() and ColoredMap.plot()
        """
        with ColoredMap(**find_args([plt.savefig, ColoredMap], **kwargs)) as cm:
            cm.title = title
            cm.directory = self._dirpath
            cm.plot(**find_args([gpd.GeoDataFrame.plot, ColoredMap.plot], **kwargs))

    def _colored_map_global(self, variable, title, date, **kwargs):
        """
        Create global colored map to show the values at country level.

        Args:
            variable (str): variable name to show
            title (str): title of the figure
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()
        """
        df = self._cleaned_df.copy()
        # Check variable name
        if variable not in df.columns:
            candidates = [col for col in df.columns if col not in self.AREA_ABBR_COLS]
            raise UnExpectedValueError(name="variable", value=variable, candidates=candidates)
        # Remove cruise ships
        df = df.loc[df[self.COUNTRY] != self.OTHERS]
        # Recognize province as a region/country
        if self.PROVINCE in df:
            with contextlib.suppress(ValueError):
                df[self.ISO3] = df[self.ISO3].cat.add_categories(["GRL"])
                df[self.COUNTRY] = df[self.COUNTRY].cat.add_categories(["Greenland"])
                df.loc[df[self.PROVINCE] == "Greenland", self.AREA_ABBR_COLS] = ["GRL", "Greenland", self.NA]
        # Select country level data
        if self.PROVINCE in df.columns:
            df = df.loc[df[self.PROVINCE] == self.NA]
        # Select date
        if date is not None:
            self._ensure_dataframe(df, name="cleaned dataset", columns=[self.DATE])
            df = df.loc[df[self.DATE] == pd.to_datetime(date)]
        df[self.COUNTRY] = df[self.COUNTRY].astype(str)
        df = df.groupby(self.COUNTRY).last().reset_index()
        # Plotting
        df.rename(columns={variable: "Value"}, inplace=True)
        self._colored_map(title=title, data=df, level=self.COUNTRY, **kwargs)

    def _colored_map_country(self, country, variable, title, date, **kwargs):
        """
        Create country-specific colored map to show the values at province level.

        Args:
            country (str): country name
            variable (str): variable name to show
            title (str): title of the figure
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of covsirphy.ColoredMap() and covsirphy.ColoredMap.plot()
        """
        df = self._cleaned_df.copy()
        iso3 = self.ensure_country_name(country)
        # Check variable name
        if variable not in df.columns:
            candidates = [col for col in df.columns if col not in self.AREA_ABBR_COLS]
            raise UnExpectedValueError(name="variable", value=variable, candidates=candidates)
        # Select country-specific data
        self._ensure_dataframe(df, name="cleaned dataset", columns=[self.ISO3, self.PROVINCE])
        df = df.loc[df[self.ISO3] == iso3]
        df = df.loc[df[self.PROVINCE] != self.NA]
        if df.empty:
            raise SubsetNotFoundError(
                country=country, country_alias=iso3, details="at province level")
        # Select date
        if date is not None:
            self._ensure_dataframe(df, name="cleaned dataset", columns=[self.DATE])
            df = df.loc[df[self.DATE] == pd.to_datetime(date)]
        df = df.groupby(self.PROVINCE).last().reset_index()
        # Plotting
        df[self.COUNTRY] = country
        df.rename(columns={variable: "Value"}, inplace=True)
        self._colored_map(title=title, data=df, level=self.PROVINCE, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.term import Term
from covsirphy.downloading.db import _DataBase


class _CSJapan(_DataBase):
    """
    Access "COVID-19 Dataset in Japan - CovsirPhy project" server.
    https://github.com/lisphilar/covid19-sir/tree/master/data
    """
    GITHUB_URL = "https://raw.githubusercontent.com"
    URL_C = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_total.csv"
    URL_P = f"{GITHUB_URL}/lisphilar/covid19-sir/master/data/japan/covid_jpn_prefecture.csv"
    # File title without extensions and suffix
    TITLE = "covid_japan"
    # Dictionary of column names
    COL_DICT = {
        "Date": Term.DATE,
        "Prefecture": Term.PROVINCE,
        "Positive": Term.C,
        "Fatal": Term.F,
        "Discharged": Term.R,
        "Tested": Term.TESTS,
    }
    # Stdout when downloading (shown at most one time)
    STDOUT = "Retrieving COVID-19 dataset from https://github.com/lisphilar/covid19-sir/data/"
    # Citation
    CITATION = "Hirokazu Takaya (2020-2022), COVID-19 dataset in Japan, GitHub repository, " \
        "https://github.com/lisphilar/covid19-sir/data/japan"
    # All columns
    _all_columns = [
        Term.DATE, Term.ISO3, Term.PROVINCE, Term.CITY,
        Term.C, Term.F, Term.R, Term.TESTS, Term.VAC, Term.VAC_BOOSTERS, Term.V_ONCE, Term.V_FULL]

    def _country(self):
        """Returns country-level data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): "Japan"
                    - Province (str): NAs
                    - City (str): NAs
                    - Confirmed (numpy.float64): the number of confirmed cases
                    - Fatal (numpy.float64): the number of fatal cases
                    - Recovered (numpy.float64): the number of recovered cases
                    - Tests (numpy.float64): the number of tests
                    - Vaccinations (numpy.float64): cumulative number of vaccinations
                    - Vaccinations_boosters (numpy.float64): cumulative number of booster vaccinations
                    - Vaccinated_once (numpy.float64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (numpy.float64): cumulative number of people who received all doses prescrived by the protocol
        """
        cols = [
            "Date", "Location", "Positive", "Tested", "Discharged", "Fatal", "Vaccinated_1st", "Vaccinated_2nd", "Vaccinated_3rd"]
        df = self._provide(url=self.URL_C, suffix="", columns=cols, date="Date", date_format="%Y-%m-%d")
        df = df.groupby("Date").sum().reset_index()
        df[self.ISO3] = "JPN"
        df[self.PROVINCE] = self.NA
        df[self.CITY] = self.NA
        df[self.V_ONCE] = df["Vaccinated_1st"].cumsum()
        df[self.V_FULL] = df["Vaccinated_2nd"].cumsum()
        df[self.VAC_BOOSTERS] = df["Vaccinated_3rd"].cumsum()
        df[self.VAC] = df[[self.V_ONCE, self.V_FULL, self.VAC_BOOSTERS]].sum(axis=1)
        return df.loc[:, self._all_columns]

    def _province(self, country):
        """Returns province-level data.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): country names
                    - Province (str): province/state/prefecture names
                    - City (str): NAs
                    - Confirmed (numpy.float64): the number of confirmed cases
                    - Fatal (numpy.float64): the number of fatal cases
                    - Recovered (numpy.float64): the number of recovered cases
                    - Population (numpy.int64): population values
                    - Tests (numpy.float64): the number of tests
        """
        if self._to_iso3(country)[0] != "JPN":
            return pd.DataFrame(columns=[
                self.DATE, self.ISO3, self.PROVINCE, self.CITY,
                self.C, self.F, self.R, self.TESTS, self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL])
        cols = ["Date", "Prefecture", "Positive", "Tested", "Discharged", "Fatal"]
        df = self._provide(url=self.URL_P, suffix="_prefecture", columns=cols, date="Date", date_format="%Y-%m-%d")
        df[self.ISO3] = "JPN"
        df[self.CITY] = self.NA
        df[self.V_ONCE] = pd.NA
        df[self.V_FULL] = pd.NA
        df[self.VAC_BOOSTERS] = pd.NA
        df[self.VAC] = pd.NA
        return df.loc[:, self._all_columns]

    def _city(self, country, province):
        """Returns city-level data.

        Args:
            country (str): country name
            province (str): province/state/prefecture name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): observation date
                    - ISO3 (str): "Japan"
                    - Province (str): province/state/prefecture names
                    - City (str): city names
                    - Confirmed (numpy.float64): the number of confirmed cases
                    - Fatal (numpy.float64): the number of fatal cases
                    - Recovered (numpy.float64): the number of recovered cases
                    - Tests (numpy.float64): the number of tests
                    - Vaccinations (numpy.float64): cumulative number of vaccinations
                    - Vaccinations_boosters (numpy.float64): cumulative number of booster vaccinations
                    - Vaccinated_once (numpy.float64): cumulative number of people who received at least one vaccine dose
                    - Vaccinated_full (numpy.float64): cumulative number of people who received all doses prescrived by the protocol
        """
        return pd.DataFrame(columns=self._all_columns)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from scipy.integrate import solve_ivp
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase


class _ODESolver(Term):
    """
    Solve initial value problems for a SIR-derived ODE model.

    Args:
        model (covsirphy.ModelBase): SIR-derived ODE model
        kwargs: values of non-dimensional model parameters, including rho and sigma

    Note:
        We can check non-dimensional model parameters with model.PARAMETERS class variable.
        All non-dimensional parameters must be specified with keyword arguments.
    """

    def __init__(self, model, **kwargs):
        self._model = Validator(model, "model").subclass(ModelBase)
        param_dict = {k: float(v) for (k, v) in kwargs.items() if isinstance(v, (float, int))}
        self._param_dict = Validator(param_dict, "kwargs").dict(required_keys=model.PARAMETERS, errors="raise")

    def run(self, step_n, **kwargs):
        """
        Solve an initial value problem.

        Args:
            step_n (int): the number of steps
            kwargs: initial values of dimensional variables, including Susceptible

        Returns:
            pandas.DataFrame: numerical solution
                Index
                    reset index: time steps
                Columns
                    (int): dimensional variables of the model

        Note:
            We can check dimensional variables with model.VARIABLES class variable.
            All dimensional variables must be specified with keyword arguments.
            Total value of initial values will be regarded as total population.
        """
        # Check arguments
        step_n = Validator(step_n, "number").int(value_range=(1, None))
        kwargs = {param: int(value) for (param, value) in kwargs.items()}
        y0_dict = Validator(kwargs, "kwargs").dict(required_keys=self._model.VARIABLES, errors="raise")
        # Calculate population
        population = sum(y0_dict.values())
        # Solve problem
        return self._run(step_n=step_n, y0_dict=y0_dict, population=population)

    def _run(self, step_n, y0_dict, population):
        """
        Solve an initial value problem for a SIR-derived ODE model.

        Args:
            step_n (int): the number of steps
            y0_dict (dict[str, int]): initial values of dimensional variables, including Susceptible
            population (int): total population

        Returns:
            pandas.DataFrame: numerical solution
                Index
                    reset index: time steps
                Columns
                    (int): dimensional variables of the model
        """
        tstart, dt, tend = 0, 1, step_n
        variables = self._model.VARIABLES[:]
        initials = [y0_dict[var] for var in variables]
        sol = solve_ivp(
            fun=self._model(population=population, **self._param_dict),
            t_span=[tstart, tend],
            y0=np.array(initials, dtype=np.int64),
            t_eval=np.arange(tstart, tend + dt, dt),
            dense_output=False
        )
        y_df = pd.DataFrame(data=sol["y"].T.copy(), columns=variables)
        return y_df.round().astype(np.int64)
